"""Defines the Data Classes used."""

import math
from datetime import datetime, timezone

# from pprint import pprint as pp

from pyastroweatherio.const import (
    CONDITION,
    CONDITION_PLAIN,
    DEEP_SKY_THRESHOLD,
    LIFTED_INDEX_PLAIN,
    LIFTED_INDEX_RANGE,
    LIFTED_INDEX_VALUE,
    MAG_DEGRATION_MAX,
    SEEING_MAX,
    WIND10M_DIRECTON,
    WIND10M_MAX,
    WIND10M_PLAIN,
    WIND10M_RANGE,
    WIND10M_VALUE,
)


class ConditionData:
    """A representation of the base class for AstroWeather Data."""

    def __init__(self, data):
        self._seventimer_init = data["seventimer_init"]
        self._seventimer_timepoint = data["seventimer_timepoint"]
        self._forecast_time = data["forecast_time"]

        self._condition_data = data["condition_data"]

    @property
    def seventimer_init(self) -> datetime:
        """Return Forecast Anchor."""

        if self._seventimer_init is not None:
            return self._seventimer_init.replace(microsecond=0, tzinfo=timezone.utc)
        return None

    @property
    def seventimer_timepoint(self) -> int:
        """Return Forecast Hour."""

        if self._seventimer_timepoint is not None:
            return self._seventimer_timepoint
        return None

    @property
    def forecast_time(self) -> datetime:
        """Return Forecast Timestamp."""

        if self._forecast_time is not None:
            return self._forecast_time.replace(microsecond=0, tzinfo=timezone.utc)
        return None

    # #########################################################################
    # Condition
    # #########################################################################
    @property
    def condition_percentage(self) -> int:
        """Return condition based on cloud cover, seeing and transparency."""

        value = self._condition_data.condition_percentage
        if value is not None:
            return int(value)
        return None

    @property
    def cloudcover(self) -> int:
        """Return Cloud Coverage."""

        value = self._condition_data.cloudcover
        if value is not None:
            return int(value)
        return None

    @property
    def cloudcover_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        value = self._condition_data.cloudcover
        if value is not None:
            return int(value)
        return None

    @property
    def cloudless_percentage(self) -> int:
        """Return Cloudless Percentage."""

        value = self._condition_data.cloudcover
        if value is not None:
            return 100 - int(value)
        return None

    @property
    def cloud_area_fraction_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        value = self._condition_data.cloud_area_fraction
        if value is not None:
            return int(value)
        return None

    @property
    def cloud_area_fraction_high_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        value = self._condition_data.cloud_area_fraction_high
        if value is not None:
            return int(value)
        return None

    @property
    def cloud_area_fraction_medium_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        value = self._condition_data.cloud_area_fraction_medium
        if value is not None:
            return int(value)
        return None

    @property
    def cloud_area_fraction_low_percentage(self) -> int:
        """Return Cloud Cover Percentage."""

        value = self._condition_data.cloud_area_fraction_low
        if value is not None:
            return int(value)
        return None

    @property
    def fog_area_fraction_percentage(self) -> int:
        """Return Fog Area Percentage."""

        value = self._condition_data.fog_area_fraction
        if value is not None:
            return int(value)
        return None

    @property
    def seeing(self) -> float:
        """Return Seeing."""

        value = self._condition_data.seeing
        if value is not None:
            return round(value, 2)
        return None

    @property
    def seeing_percentage(self) -> int:
        """Return Seeing."""

        value = self._condition_data.seeing
        if value is not None:
            return int(100 - value * 100 / SEEING_MAX)
        return None

    @property
    def transparency(self) -> float:
        """Return Transparency."""

        value = self._condition_data.transparency
        if value is not None:
            return round(value, 2)
        return None

    @property
    def transparency_percentage(self) -> int:
        """Return Transparency."""

        value = self._condition_data.transparency
        if value is not None:
            return int(100 - value * 100 / MAG_DEGRATION_MAX)
        return None

    @property
    def lifted_index(self) -> float:
        """Return Lifted Index."""

        value = self._condition_data.lifted_index
        if value is not None:
            return round(value, 2)
        return None

    @property
    def rh2m(self) -> int:
        """Return 2m Relative Humidity."""

        value = self._condition_data.rh2m
        if value is not None:
            return value
        return None

    @property
    def wind10m_speed(self) -> float:
        """Return 10m Wind Speed."""

        value = self._condition_data.wind_speed
        if value is not None:
            return value
        return None

    @property
    def calm_percentage(self) -> int:
        """Return 10m Wind Speed."""

        value = self._condition_data.wind_speed
        if value is not None:
            return int(100 - value * (100 / WIND10M_MAX))
        return None

    @property
    def wind10m_direction(self) -> str:
        """Return 10m Wind Direction."""

        value = self._condition_data.wind_from_direction
        if value is not None:
            direction = value
            direction += 22.5
            direction = direction % 360
            direction = int(direction / 45)  # values 0 to 7
            return WIND10M_DIRECTON[max(0, min(7, direction))]
        return None

    @property
    def temp2m(self) -> int:
        """Return 2m Temperature."""

        value = self._condition_data.temp2m
        if value is not None:
            return value
        return None

    @property
    def dewpoint2m(self) -> float:
        """Return 2m Dew Point."""

        value = self._condition_data.dewpoint2m
        if value is not None:
            return round(value, 1)
        return None

    @property
    def weather(self) -> str:
        """Return Current Weather."""

        value = self._condition_data.weather
        if value is not None:
            return value.replace("_", " ").capitalize()
        return None

    @property
    def weather6(self) -> str:
        """Return Current Weather."""

        value = self._condition_data.weather6
        if value is not None:
            return value.replace("_", " ").capitalize()
        return None

    @property
    def precipitation_amount(self) -> float:
        """Return Current Precipitation Amount."""

        value = self._condition_data.precipitation_amount
        if value is not None:
            return value
        return None

    @property
    def precipitation_amount6(self) -> float:
        """Return Precipitation Amount in next 6 hours."""

        value = self._condition_data.precipitation_amount6
        if value is not None:
            return value
        return None


class LocationData(ConditionData):
    """A representation of the Location AstroWeather Data."""

    def __init__(self, data):
        super().__init__(data)
        self._time_shift = data["time_shift"]
        self._forecast_length = data["forecast_length"]

        self._location_data = data["location_data"]
        self._sun_data = data["sun_data"]
        self._moon_data = data["moon_data"]
        self._darkness_data = data["darkness_data"]
        self._night_duration_astronomical = data["night_duration_astronomical"]
        self._deepsky_forecast = data["deepsky_forecast"]
        self._uptonight = data["uptonight"]
        self._uptonight_bodies = data["uptonight_bodies"]
        self._uptonight_comets = data["uptonight_comets"]

    # #########################################################################
    # Location
    # #########################################################################
    @property
    def time_shift(self) -> int:
        """Return Forecast Timestamp."""

        if self._time_shift is not None:
            return self._time_shift
        return None

    @property
    def forecast_length(self) -> int:
        """Return Forecast Length in Hours."""

        if self._forecast_length is not None:
            return self._forecast_length
        return None

    @property
    def latitude(self) -> float:
        """Return Latitude."""

        if self._location_data.latitude is not None:
            return self._location_data.latitude
        return None

    @property
    def longitude(self) -> float:
        """Return Longitude."""

        if self._location_data.longitude is not None:
            return self._location_data.longitude
        return None

    @property
    def elevation(self) -> float:
        """Return Elevation."""

        if self._location_data.elevation is not None:
            return self._location_data.elevation
        return None

    # #########################################################################
    # Condition
    # #########################################################################
    @property
    def seeing_plain(self) -> str:
        """Return Seeing."""

        return "Deprecated. Use seeing instead."

    @property
    def wind10m_speed_plain(self) -> str:
        """Return wind speed plain."""

        value = self._condition_data.wind_speed
        if value is not None:
            wind_speed_value = 0
            for (start, end), derate in zip(WIND10M_RANGE, WIND10M_VALUE):
                if start <= value <= end:
                    wind_speed_value = derate

            return WIND10M_PLAIN[max(0, min(7, wind_speed_value - 1))]
        return None

    @property
    def lifted_index_plain(self) -> str:
        """Return Lifted Index plain."""

        value = self._condition_data.lifted_index
        if value is not None:
            lifted_index_value = 0
            for (start, end), derate in zip(LIFTED_INDEX_RANGE, LIFTED_INDEX_VALUE):
                if start <= value <= end:
                    lifted_index_value = derate

            return LIFTED_INDEX_PLAIN[max(0, min(7, lifted_index_value - 1))]
        return None

    @property
    def deep_sky_view(self) -> bool:
        """Return True if Deep Sky should be possible."""

        if self.condition_percentage is not None:
            if self.condition_percentage >= DEEP_SKY_THRESHOLD:
                return True
            return False
        return None

    @property
    def condition_plain(self) -> str:
        """Return Current View Conditions."""

        if self.condition_percentage is not None:
            if self.condition_percentage > 80:
                return CONDITION_PLAIN[0].capitalize()
            if self.condition_percentage > 60:
                return CONDITION_PLAIN[1].capitalize()
            if self.condition_percentage > 40:
                return CONDITION_PLAIN[2].capitalize()
            if self.condition_percentage > 20:
                return CONDITION_PLAIN[3].capitalize()
            return CONDITION_PLAIN[4].capitalize()
        return None

    # #########################################################################
    # Sun
    # #########################################################################
    @property
    def sun_altitude(self) -> float:
        """Return Sun Altitude."""

        value = self._sun_data.altitude
        if value is not None:
            return round(value, 3)
        return None

    @property
    def sun_azimuth(self) -> float:
        """Return sun Azimuth."""

        value = self._sun_data.azimuth
        if value is not None:
            return round(value, 3)
        return None

    @property
    def sun_next_rising_astro(self) -> datetime:
        """Return Sun Next Rising Astronomical."""

        value = self._sun_data.next_rising_astro
        if value is not None:
            return value
        return None

    @property
    def sun_next_rising(self) -> datetime:
        """Return Sun Next Rising Civil."""

        value = self._sun_data.next_rising_civil
        if value is not None:
            return value
        return None

    @property
    def sun_next_rising_nautical(self) -> datetime:
        """Return Sun Next Rising Nautical."""

        value = self._sun_data.next_rising_nautical
        if value is not None:
            return value
        return None

    @property
    def sun_next_setting_astro(self) -> datetime:
        """Return Sun Next Setting Astronomical."""

        value = self._sun_data.next_setting_astro
        if value is not None:
            return value
        return None

    @property
    def sun_next_setting(self) -> datetime:
        """Return Next Setting Civil."""

        value = self._sun_data.next_setting_civil
        if value is not None:
            return value
        return None

    @property
    def sun_next_setting_nautical(self) -> datetime:
        """Return Sun Next Setting Nautical."""

        value = self._sun_data.next_setting_nautical
        if value is not None:
            return value
        return None

    # #########################################################################
    # Moon
    # #########################################################################
    @property
    def moon_altitude(self) -> float:
        """Return Moon Altitude."""

        value = self._moon_data.altitude
        if value is not None:
            return round(value, 3)
        return None

    @property
    def moon_angular_size(self) -> float:
        """Return Moon Angular Size in Minutes."""

        value = self._moon_data.angular_size
        if value is not None:
            return round(value, 3)
        return None

    @property
    def moon_azimuth(self) -> float:
        """Return Moon Azimuth."""

        value = self._moon_data.azimuth
        if value is not None:
            return round(value, 3)
        return None

    @property
    def moon_distance_km(self) -> float:
        """Return Moon Distance in km."""

        value = self._moon_data.distance_km
        if value is not None:
            return round(value, 0)
        return None

    @property
    def moon_next_full_moon(self) -> datetime:
        """Return Moon Next Full Moon."""

        value = self._moon_data.next_full_moon
        if value is not None:
            return value
        return None

    @property
    def moon_next_new_moon(self) -> datetime:
        """Return Moon Next New Moon."""

        value = self._moon_data.next_new_moon
        if value is not None:
            return value
        return None

    @property
    def moon_next_rising(self) -> datetime:
        """Return Moon Next Rising."""

        value = self._moon_data.next_rising
        if value is not None:
            return value
        return None

    @property
    def moon_next_setting(self) -> datetime:
        """Return Moon Next Setting."""

        value = self._moon_data.next_setting
        if value is not None:
            return value
        return None

    @property
    def moon_phase(self) -> float:
        """Return Moon Phase."""

        value = self._moon_data.phase
        if value is not None:
            return round(value, 1)
        return None

    @property
    def moon_relative_size(self) -> float:
        """Return Moon Relative Size in %."""

        value = self._moon_data.relative_size
        if value is not None:
            return round(value * 100 - 100, 3)
        return None

    # #########################################################################
    # Darkness
    # #########################################################################
    @property
    def night_duration_astronomical(self) -> float:
        """Returns the remaining timespan of astronomical darkness."""

        if self._night_duration_astronomical is not None:
            return self._night_duration_astronomical
        return None

    @property
    def deep_sky_darkness_moon_rises(self) -> bool:
        """Returns true if moon rises during astronomical night."""

        return self._darkness_data.deep_sky_darkness_moon_rises

    @property
    def deep_sky_darkness_moon_sets(self) -> bool:
        """Returns true if moon sets during astronomical night."""

        return self._darkness_data.deep_sky_darkness_moon_sets

    @property
    def deep_sky_darkness_moon_always_up(self) -> bool:
        """Returns true if moon is up during astronomical night."""

        return self._darkness_data.deep_sky_darkness_moon_always_up

    @property
    def deep_sky_darkness_moon_always_down(self) -> bool:
        """Returns true if moon is down during astronomical night."""

        return self._darkness_data.deep_sky_darkness_moon_always_down

    @property
    def deep_sky_darkness(self) -> float:
        """Returns the remaining timespan of deep sky darkness."""

        return self._darkness_data.deep_sky_darkness

    # #########################################################################
    # Deep Sky Forecast
    # #########################################################################
    @property
    def deepsky_forecast_today(self) -> int:
        """Return Forecas Today in Percent."""

        if self._deepsky_forecast is not None:
            nightly_condition_sum = 0
            if len(self._deepsky_forecast) > 0:
                for nightly_condition in self._deepsky_forecast[0].nightly_conditions:
                    nightly_condition_sum += nightly_condition
                return int(round(nightly_condition_sum / len(self._deepsky_forecast[0].nightly_conditions)))
        return None

    @property
    def deepsky_forecast_today_dayname(self):
        """Return Forecast Todays Dayname."""

        if self._deepsky_forecast is not None:
            if len(self._deepsky_forecast) > 0:
                nightly_conditions = self._deepsky_forecast[0]
                return nightly_conditions.dayname
        return None

    @property
    def deepsky_forecast_today_plain(self):
        """Return Forecast Today."""

        if self._deepsky_forecast is not None:
            out = ""
            if len(self._deepsky_forecast) > 0:
                for nightly_condition in self._deepsky_forecast[0].nightly_conditions:
                    out += CONDITION[4 - math.floor(nightly_condition / 20)].capitalize()
            return out
        return None

    @property
    def deepsky_forecast_today_desc(self):
        """Return Forecast Today Description."""

        if self._deepsky_forecast is not None:
            if len(self._deepsky_forecast) > 0:
                nightly_conditions = self._deepsky_forecast[0]
                return nightly_conditions.weather.replace("_", " ").capitalize()
        return None

    @property
    def deepsky_forecast_today_precipitation_amount6(self) -> float:
        """Return Forecast Today Precipitation Amount6."""

        if self._deepsky_forecast is not None:
            if len(self._deepsky_forecast) > 0:
                nightly_conditions = self._deepsky_forecast[0]
                return nightly_conditions.precipitation_amount6
        return None

    @property
    def deepsky_forecast_tomorrow(self) -> int:
        """Return Forecas Tomorrow in Percentt."""

        if self._deepsky_forecast is not None:
            nightly_condition_sum = 0
            if len(self._deepsky_forecast) > 1:
                for nightly_condition in self._deepsky_forecast[1].nightly_conditions:
                    nightly_condition_sum += nightly_condition
                return int(round(nightly_condition_sum / len(self._deepsky_forecast[1].nightly_conditions)))
        return None

    @property
    def deepsky_forecast_tomorrow_dayname(self):
        """Return Forecast Todays Dayname."""

        if self._deepsky_forecast is not None:
            if len(self._deepsky_forecast) > 1:
                nightly_conditions = self._deepsky_forecast[1]
                return nightly_conditions.dayname
        return None

    @property
    def deepsky_forecast_tomorrow_plain(self):
        """Return Forecast Tomorrow."""

        if self._deepsky_forecast is not None:
            out = ""
            if len(self._deepsky_forecast) > 1:
                for nightly_condition in self._deepsky_forecast[1].nightly_conditions:
                    out += CONDITION[4 - math.floor(nightly_condition / 20)].capitalize()
            return out
        return None

    @property
    def deepsky_forecast_tomorrow_desc(self):
        """Return Forecast Tomorrow Description."""

        if self._deepsky_forecast is not None:
            if len(self._deepsky_forecast) > 1:
                nightly_conditions = self._deepsky_forecast[1]
                return nightly_conditions.weather.replace("_", " ").capitalize()
        return None

    @property
    def deepsky_forecast_tomorrow_precipitation_amount6(self) -> float:
        """Return Forecast Today Precipitation Amount6."""

        if self._deepsky_forecast is not None:
            if len(self._deepsky_forecast) > 0:
                nightly_conditions = self._deepsky_forecast[1]
                return nightly_conditions.precipitation_amount6
        return None

    @property
    def deepsky_forecast(self):
        """Return Deepsky Forecast."""

        if self._deepsky_forecast is not None:
            return self._deepsky_forecast
        return None

    # #########################################################################
    # UpTonight
    # #########################################################################
    @property
    def uptonight(self) -> int:
        """Return the number of best DSOs for tonight."""

        if self._uptonight is not None:
            return len(self._uptonight)
        return None

    @property
    def uptonight_list(self) -> []:
        """Return the list of UpTonight targets."""

        if self._uptonight is not None:
            return self._uptonight
        return None

    # #########################################################################
    # UpTonight Bodies
    # #########################################################################
    @property
    def uptonight_bodies(self) -> int:
        """Return the number of best BODIEs for tonight."""

        if self._uptonight_bodies is not None:
            return len(self._uptonight_bodies)
        return None

    @property
    def uptonight_bodies_list(self) -> []:
        """Return the list of UpTonight bodies."""

        if self._uptonight_bodies is not None:
            return self._uptonight_bodies
        return None

    # #########################################################################
    # UpTonight Comets
    # #########################################################################
    @property
    def uptonight_comets(self) -> int:
        """Return the number of best comets for tonight."""

        if self._uptonight_comets is not None:
            return len(self._uptonight_comets)
        return None

    @property
    def uptonight_comets_list(self) -> []:
        """Return the list of UpTonight comets."""

        if self._uptonight_comets is not None:
            return self._uptonight_comets
        return None


class ForecastData(ConditionData):
    """A representation of 3-Hour Based Forecast AstroWeather Data."""

    def __init__(self, data):
        super().__init__(data)
        self._hour = data["hour"]

    @property
    def hour(self) -> int:
        """Return Forecast Hour of the day."""

        if self._hour is not None:
            return self._hour
        return None

    @property
    def deep_sky_view(self) -> bool:
        """Return True if Deep Sky should be possible."""

        if self.condition_percentage is not None:
            if self.condition_percentage <= DEEP_SKY_THRESHOLD:
                return True
            return False
        return None


class NightlyConditionsData:
    """A representation of nights Sky Quality Data."""

    def __init__(self, data) -> None:
        self._seventimer_init = data["seventimer_init"]
        self._dayname = data["dayname"]
        self._hour = data["hour"]
        self._nightly_conditions = data["nightly_conditions"]
        self._weather = data["weather"]
        self._precipitation_amount6 = data["precipitation_amount6"]

    @property
    def seventimer_init(self) -> datetime:
        """Return Forecast Anchor."""

        if self._seventimer_init is not None:
            return self._seventimer_init
        return None

    @property
    def dayname(self) -> str:
        """Return Forecast Name of the Day."""

        if self._dayname is not None:
            return self._dayname
        return None

    @property
    def hour(self) -> int:
        """Return Forecast Hour."""

        if self._hour is not None:
            return self._hour
        return None

    @property
    def nightly_conditions(self) -> int:
        """Return Forecast Hour."""

        if self._nightly_conditions is not None:
            return self._nightly_conditions
        return None

    @property
    def weather(self) -> str:
        """Return Current Weather."""

        if self._weather is not None:
            return self._weather.replace("_", " ").capitalize()
        return None

    @property
    def precipitation_amount6(self) -> float:
        """Return Current Precipitation Amount 6hrs."""

        if self._precipitation_amount6 is not None:
            return self._precipitation_amount6
        return None


class DSOUpTonight:
    """A representation of uptonight DSO."""

    def __init__(self, data):
        self._id = data["id"]
        self._target_name = data["target_name"]
        self._type = data["type"]
        self._constellation = data["constellation"]
        self._size = data["size"]
        self._visual_magnitude = data["visual_magnitude"]
        self._meridian_transit = data["meridian_transit"]
        self._meridian_antitransit = data["meridian_antitransit"]
        self._foto = data["foto"]

    @property
    def id(self) -> str:
        """Return object catalogue id."""

        if self._id is not None:
            return self._id
        return None

    @property
    def target_name(self) -> str:
        """Return object name."""

        if self._target_name is not None:
            return self._target_name
        return None

    @property
    def type(self) -> str:
        """Return the type of the object."""

        if self._type is not None:
            return self._type
        return None

    @property
    def constellation(self) -> str:
        """Return the constellation of the object."""

        if self._constellation is not None:
            return self._constellation
        return None

    @property
    def size(self) -> float:
        """Return size of the object."""

        if self._size is not None:
            return self._size
        return None

    @property
    def visual_magnitude(self) -> float:
        """Return visual magnitude of the object."""

        if self._visual_magnitude is not None:
            return self._visual_magnitude
        return None

    @property
    def meridian_transit(self) -> datetime:
        """Return the meridian transit time of the object."""

        if self._meridian_transit is not None:
            return self._meridian_transit
        return None

    @property
    def meridian_antitransit(self) -> datetime:
        """Return the meridian anti-transit time of the object."""

        if self._meridian_antitransit is not None:
            return self._meridian_antitransit
        return None

    @property
    def foto(self) -> float:
        """Return fraction of time observable of the object."""

        if self._foto is not None:
            return self._foto
        return None


class BODIESUpTonight:
    """A representation of uptonight bodies."""

    def __init__(self, data):
        self._target_name = data["target_name"]
        self._max_altitude = data["max_altitude"]
        self._azimuth = data["azimuth"]
        self._max_altitude_time = data["max_altitude_time"]
        self._visual_magnitude = data["visual_magnitude"]
        self._meridian_transit = data["meridian_transit"]
        self._foto = data["foto"]

    @property
    def target_name(self) -> str:
        """Return body name."""

        if self._target_name is not None:
            return self._target_name
        return None

    @property
    def max_altitude(self) -> float:
        """Return maximum altitude of the body."""

        if self._max_altitude is not None:
            return self._max_altitude
        return None

    @property
    def azimuth(self) -> float:
        """Return the azimuth of maximum altitude."""

        if self._azimuth is not None:
            return self._azimuth
        return None

    @property
    def max_altitude_time(self) -> datetime:
        """Return date and time of maximum altitude."""

        if self._max_altitude_time is not None:
            return self._max_altitude_time
        return None

    @property
    def visual_magnitude(self) -> float:
        """Return visual magnitude of the body."""

        if self._visual_magnitude is not None:
            return self._visual_magnitude
        return None

    @property
    def meridian_transit(self) -> datetime:
        """Return the meridian transit time of the body."""

        if self._meridian_transit is not None:
            return self._meridian_transit
        return None

    @property
    def foto(self) -> float:
        """Return fraction of time observable of the body."""

        if self._foto is not None:
            return self._foto
        return None


class COMETSUpTonight:
    """A representation of uptonight comets."""

    def __init__(self, data):
        self._designation = data["designation"]
        self._distance_au_earth = data["distance_au_earth"]
        self._distance_au_sun = data["distance_au_sun"]
        self._absolute_magnitude = data["absolute_magnitude"]
        self._visual_magnitude = data["visual_magnitude"]
        self._altitude = data["altitude"]
        self._azimuth = data["azimuth"]
        self._rise_time = data["rise_time"]
        self._set_time = data["set_time"]

    @property
    def designation(self) -> str:
        """Return comet name."""

        if self._designation is not None:
            return self._designation
        return None

    @property
    def distance_au_earth(self) -> float:
        """Return distance to earth in au."""

        if self._distance_au_earth is not None:
            return self._distance_au_earth
        return None

    @property
    def distance_au_sun(self) -> float:
        """Return distance to sun in au."""

        if self._distance_au_sun is not None:
            return self._distance_au_sun
        return None

    @property
    def absolute_magnitude(self) -> float:
        """Return absolute magnitude."""

        if self._absolute_magnitude is not None:
            return self._absolute_magnitude
        return None

    @property
    def visual_magnitude(self) -> float:
        """Return visual magnitude."""

        if self._visual_magnitude is not None:
            return self._visual_magnitude
        return None

    @property
    def altitude(self) -> float:
        """Return altitude."""

        if self._altitude is not None:
            return self._altitude
        return None

    @property
    def azimuth(self) -> float:
        """Return the azimuth of maximum altitude."""

        if self._azimuth is not None:
            return self._azimuth
        return None

    @property
    def rise_time(self) -> datetime:
        """Return rise time."""

        if self._rise_time is not None:
            return self._rise_time
        return None

    @property
    def set_time(self) -> datetime:
        """Return rise time."""

        if self._set_time is not None:
            return self._set_time
        return None



"""Contains Helper functions for AstroWeather."""

import logging
import math
from datetime import UTC, datetime, timedelta
from math import degrees as deg

import ephem
from ephem import degree
from pydantic import ValidationError
from zoneinfo import ZoneInfo

# import metpy.calc as mpcalc
# from metpy.units import units
# import numpy as np
from pyastroweatherio.const import (
    ASTRONOMICAL_DUSK_DAWN,
    CIVIL_DUSK_DAWN,
    MAG_DEGRATION_MAX,
    NAUTICAL_DUSK_DAWN,
    SEEING_MAX,
)
from pyastroweatherio.models import DarknessModel, MoonModel, SunModel

_LOGGER = logging.getLogger(__name__)


class ConversionFunctions:
    """Convert between different units."""

    async def epoch_to_datetime(self, value) -> str:
        """Converts EPOC time to Date Time String."""

        return datetime.datetime.fromtimestamp(int(value)).strftime("%Y-%m-%d %H:%M:%S")

    async def anchor_timestamp(self, value) -> datetime:
        """Converts the datetime string from 7Timer to DateTime."""

        return datetime.strptime(value, "%Y%m%d%H")


class AtmosphericRoutines:
    """Calculate atmospheric attributes."""

    def __init__(
        self,
    ):
        _LOGGER.debug("AtmosphericRoutines calculation mode active")

        # Relative humidity (0.0 to 1.0)
        # Temperature in Celsius
        # Cloud cover fraction (0.0 to 1.0)
        # Wind speed in meters per second
        # Humidity in percent
        # Altitude in meters
        # Dew point temperature in Celsius
        # Air pressure at sea level in hPa
        # aerosol_density in kg/m^3

    # #####################################################
    # Calculate lifted index
    # #####################################################
    async def calculate_lifted_index(self, temperature, altitude, dew_point_temperature, air_pressure_at_sea_level):
        """Calculate atmospheric lifted index."""
        # https://en.wikipedia.org/wiki/Lifted_index

        if not all(
            v is not None
            for v in [
                temperature,
                altitude,
                dew_point_temperature,
                air_pressure_at_sea_level,
            ]
        ):
            return None

        # Constants
        env_temp_500mb = -20  # Celsius (environmental temperature at 500 mb level)

        # Calculate saturation vapor pressure at surface temperature
        # Checked with https://www.weather.gov/epz/wxcalc_vaporpressure
        es = self._calculate_vapor_pressure(temperature)

        # Calculate actual Vapor Pressure at surface
        # Checked with https://www.weather.gov/epz/wxcalc_vaporpressure
        e = self._calculate_vapor_pressure(dew_point_temperature)  # 6.112 * (10 ** (7.5 * (Td - Tn) / (Td - 35.85)))

        # Calculate Mixing Ratio at Surface in grams per kilogram
        # Checked with https://www.weather.gov/epz/wxcalc_mixingratio
        w = self._calculate_mixing_ratio(e, air_pressure_at_sea_level)

        # Calculate Lifting Condensation Level
        lcl = self._calculate_lifting_condensation_level(w, air_pressure_at_sea_level)

        # Calculate temperature of lifted parcel at 500 mb level
        lifted_temp_500mb = (
            env_temp_500mb + (temperature - lcl) * 0.5
        )  # Assumption: 500 mb is halfway through the troposphere

        # Calculate Lifted Index
        lifted_index = (env_temp_500mb - lifted_temp_500mb) * 2 + 7

        # Ensure lifted index is within the valid range [-7, 7]
        lifted_index = max(-7, min(7, lifted_index))

        # _LOGGER.debug(f"Vapor Pressure: {e}")
        # _LOGGER.debug(f"Mixing Ratio: {w}")
        # _LOGGER.debug(f"Lifting Condensation Level: {lcl}")
        # _LOGGER.debug(f"Temperature of Lifted Parcel: {lifted_temp_500mb}")
        # _LOGGER.debug(f"Lifted Index: {lifted_index}")

        return lifted_index

    # #####################################################
    # Calculate magniture degradation based on transparency
    # #####################################################
    async def magnitude_degradation(
        self,
        temperature,
        humidity,
        cloud_cover,
        wind_speed,
        altitude,
        dew_point_temperature,
        air_pressure_at_sea_level,
    ):
        """
        Calculates the magnitude_degradation of the atmosphere.
        This algorithm first calculates the lifted index and the seeing and uses them to calculate
        the transparency from which the magnitude degradation is derived from.

        Args:
        - temperature: Surface temperature in Celsius.
        - humidity: Humidity in Percent.
        - dew_point_temperature: Dew point temperature in Celsius.
        - wind_speed: Wind speed in meters per second.
        - altitude: Altitude in meters.
        - air_pressure_at_sea_level: Air pressure at sea level.
        - cloud_cover: Cloud cover in Percent.
        - lifted_index:
        - seeing:

        Returns:
        - seeing: In Arcsecs
        """

        if not all(
            v is not None
            for v in [
                temperature,
                humidity,
                cloud_cover,
                wind_speed,
                altitude,
                dew_point_temperature,
                air_pressure_at_sea_level,
            ]
        ):
            return None

        lifted_index = await self.calculate_lifted_index(
            temperature, altitude, dew_point_temperature, air_pressure_at_sea_level
        )
        seeing = await self.calculate_seeing(
            temperature,
            humidity,
            dew_point_temperature,
            wind_speed,
            cloud_cover,
            altitude,
            air_pressure_at_sea_level,
        )

        # Calculate transparency
        transparency = self._calculate_transparency(
            humidity,
            temperature,
            cloud_cover,
            wind_speed,
            altitude,
            dew_point_temperature,
            air_pressure_at_sea_level,
            lifted_index,
            seeing,
        )

        # Convert transparency to magnitude degradation
        magnitude_degradation = self._transparency_to_magnitude_degradation(transparency)

        # _LOGGER.debug(
        #     "Magnitude Degradation: {:.2f} mag (".format(magnitude_degradation)
        #     + "Lifted Index (LI): {:.2f} °C, ".format(lifted_index)
        #     + "Seeing: {:.2f} arcsec, ".format(seeing)
        #     + "Estimated Atmospheric Transparency: {:.2f})".format(transparency)
        # )

        return magnitude_degradation

    # #####################################################
    # Calculate atmospheric seeing
    # #####################################################
    # Modell 6:
    # This algorithm first calculates the seeing factor based on temperature, humidity, wind speed
    # and altitude above sea level. It is similar to Model 4 and 5 but uses the air pressure at sea level
    # and dew point provided by Met.no.
    async def calculate_seeing(
        self,
        temperature,
        humidity,
        dew_point_temperature,
        wind_speed,
        cloud_cover,
        altitude,
        air_pressure_at_sea_level,
    ):
        """
        Calculated seeing of the atmosphere. This algorithm first calculates the seeing factor based on temperature,
        humidity, wind speed and altitude above sea level. The seeing factor is then used to calculate the astronomical
        seeing in arcseconds. The empirical relationship used here states that the seeing in arcseconds is approximately
        equal to the reciprocal of the seeing factor multiplied by a conversion factor of 0.98.

        Used by: magnitude degradation

        Args:
        - temperature: Surface temperature in Celsius.
        - humidity: Humidity in Percent.
        - dew_point_temperature: Dew point temperature in Celsius.
        - wind_speed: Wind speed in meters per second.
        - cloud_cover: Cloud cover in Percent.
        - altitude: Altitude in meters.
        - air_pressure_at_sea_level: Air pressure at sea level.

        Returns:
        - seeing: In Arcsecs

        Flow:
        - _calculate_water_vapor_pressure
        - adjusted_pressure
        - relative_pressure
        - seeing_factor
        - seeing
        """

        if not all(
            v is not None
            for v in [
                temperature,
                humidity,
                dew_point_temperature,
                wind_speed,
                cloud_cover,
                altitude,
                air_pressure_at_sea_level,
            ]
        ):
            return None

        # Constants
        C = 6.5  # 1.7

        water_vapor_pressure = self._calculate_water_vapor_pressure(dew_point_temperature, humidity)

        # adjusted_pressure = air_pressure_at_sea_level * math.exp(-0.00012 * altitude)
        adjusted_pressure = self._calculate_adjusted_pressure(air_pressure_at_sea_level, altitude)
        relative_pressure = adjusted_pressure / air_pressure_at_sea_level

        seeing_factor = C * (water_vapor_pressure / 10) ** 0.25 * (wind_speed / 10) ** 0.75 * relative_pressure
        seeing = 0.98 / seeing_factor

        # _LOGGER.debug(
        #     "Seeing: {:.2f} arcsec (".format(seeing)
        #     + "Water Vapor Pressure: {:.2f} mbar, ".format(water_vapor_pressure)
        #     + "Wind Speed: {:.2f} m/s, ".format(wind_speed)
        #     + "Relative Pressure: {:.2f} mbar, ".format(relative_pressure)
        #     + "Seeing Factor: {:.2f})".format(seeing_factor)
        # )

        if seeing > SEEING_MAX:
            seeing = SEEING_MAX  # max out seeing

        return seeing

    # #####################################################
    # Atmospheric calculations
    # #####################################################
    def _calculate_adjusted_pressure(self, pressure_sea_level, altitude):
        """
        Calculate the adjusted pressure at a given altitude above sea level.
        """

        lapse_rate = -0.0065  # Temperature lapse rate in K/m
        temperature_sea_level = 288.15  # Temperature at sea level in K
        gravity = 9.80665  # Acceleration due to gravity in m/s^2
        molar_mass_air = 0.02896  # Molar mass of Earth's air in kg/mol
        gas_constant = 8.31447  # Universal gas constant in J/(mol*K)

        pressure_adjusted = pressure_sea_level * (1 - (lapse_rate * altitude) / temperature_sea_level) ** (
            (gravity * molar_mass_air) / (gas_constant * lapse_rate)
        )

        return pressure_adjusted

    def _calculate_vapor_pressure(self, temperature):
        """
        Calculate the actual or saturation vapor pressure at the surface using the Magnus-Tetens formula.
        If the surface temperature is given, the actual vapor pressure is calculated.
        The dew point temperature calculates ths saturation vapor pressure.

        Used by: lifted index, seeing

        Args:
        - temperature: Surface or dew point temperature in Celsius.

        Returns:
        - e: Actual vapor pressure at the surface in millibars (mb) or hectopascals (hPa).

        Magnus coefficients:

        Temperatures in between -45°C bis 60°C:
            a=17.62
            b=243.12
        Temperatures in between 0°C bis 100°C:
            a=7.5
            b=237.7
        Temperatures in between -10°C bis 50°C:
            a=17.27
            b=237.7
        """
        # # Constants
        # A = 17.62
        # B = 243.12

        # # Monteith and Unsworth (2008) provide Tetens' formula for temperatures above 0 °C
        A = 17.27
        B = 237.3

        # Murray (1967) provides Tetens' equation for temperatures below 0 °C
        # A = 21.875
        # B = 265.5

        # Calculate vapor pressure using Magnus-Tetens formula
        e = 0.61078 * math.exp((A * temperature) / (temperature + B))

        return e

    def _calculate_water_vapor_pressure(self, dew_point_temperature, humidity):
        """
        Calculate the water vapor pressure based on temperature and humiditye.

        Used by: seeing

        Args:
        - dew_point_temperature: Dew point temperature in Celsius.
        - humidity: Humidity in Percent.

        Returns:
        - water_vapor_pressure: Water vapor pressure at the surface in millibars (mb) or hectopascals (hPa).
        """

        es = self._calculate_vapor_pressure(dew_point_temperature)
        water_vapor_pressure = (humidity / 100) * es

        return water_vapor_pressure

    def _calculate_mixing_ratio(self, e, air_pressure_at_sea_level):
        """
        Calculate the mixing ratio.

        Used by: lifted index

        Args:
        - e: Vapor pressure
        - air_pressure_at_sea_level: Air pressure at sea level

        Returns:
        - w: mixing ratio at surface in grams per kilogram
        """

        # Constants
        A = 621.97

        # Calculate actual vapor pressure using Magnus-Tetens formula
        w = A * (e / (air_pressure_at_sea_level - e))

        return w

    def _calculate_lifting_condensation_level(self, w, air_pressure_at_sea_level):
        """
        The Lifting Condensation Level is the level at which a parcel becomes saturated. It can be used as
        a reasonable estimate of cloud base height when parcels experience forced ascent.
        Here, we calculate it using the Clausius-Clapeyron equation. This equation relates the saturation vapor
        pressure (e) to the temperature and pressure.
        https://en.wikipedia.org/wiki/Lifted_condensation_level

        Used by: lifted index

        Args:
        - w: Mixing ratio
        - air_pressure_at_sea_level: Air pressure at sea level

        Returns:
        - lcl: LCL in meters
        """

        # Constants
        A = 2440
        B = 0.00029

        # Calculate Lifting Condensation Level using the Clausius-Clapeyron equation
        lcl = (A * w) / ((air_pressure_at_sea_level - w) * (1 - B * air_pressure_at_sea_level))

        return lcl

    def _calculate_transparency(
        self,
        humidity,
        temperature,
        cloud_cover,
        wind_speed,
        altitude,
        dew_point_temperature,
        air_pressure_at_sea_level,
        lifted_index,
        seeing,
    ):
        """
        The transparency of the atmosphere calculated by a weighted sum of conditions.

        Used by: magnitude degradation

        Args:
        - humidity: Humidity in Percent.
        - temperature: Surface temperature in Celsius.
        - cloud_cover: Cloud cover in Percent.
        - wind_speed: Wind speed in meters per second.
        - altitude: Altitude in meters.
        - dew_point_temperature: Dew point temperature in Celsius.
        - air_pressure_at_sea_level: Air pressure at sea level.
        - lifted_index: Lifted index in Celsius.
        - seeing: Seeing in Arcsecs.
        - w: Mixing ratio

        Returns:
        - transparency: In the range of 0 to 1
        """

        # Coefficients for the linear model (you can adjust these based on your requirements)
        temp_weight = 0.10
        humidity_weight = 0.10
        wind_weight = 0.10
        dew_point_weight = 0.10
        lifted_index_weight = 0.15
        seeing_weight = 0.15
        pressure_weight = 0.10
        cloud_cover_weight = 0.10
        altitude_weight = 0.10

        # Normalize values (optional)
        # Assuming typical temperature range from -10°C to 40°C
        temp_normalized = (temperature - (-10)) / (40 - (-10))
        humidity_normalized = humidity / 100.0
        # Assuming maximum typical wind speed as 20 m/s
        wind_normalized = wind_speed / 20.0
        # Assuming typical dew point range from -10°C to 30°C
        dew_point_normalized = (dew_point_temperature - (-10)) / (30 - (-10))
        # Assuming lifted index range from -8 to 8
        lifted_index_normalized = (lifted_index + 8) / 16.0
        # Assuming typical seeing conditions from 0 to 10 arcseconds
        seeing_normalized = (10 - seeing) / 10.0
        # Assuming typical pressure range from 900 to 1100 mb
        pressure_normalized = (air_pressure_at_sea_level - 900) / (1100 - 900)
        cloud_cover_normalized = cloud_cover / 100.0
        # Assuming typical altitude range from 0 to 5000 meters
        altitude_normalized = altitude / 5000.0

        transparency = 1 - (
            temp_weight * temp_normalized
            + humidity_weight * humidity_normalized
            + wind_weight * wind_normalized
            + dew_point_weight * dew_point_normalized
            + lifted_index_weight * lifted_index_normalized
            + seeing_weight * seeing_normalized
            + pressure_weight * pressure_normalized
            + cloud_cover_weight * cloud_cover_normalized
            + altitude_weight * altitude_normalized
        )

        # Ensure transparency is within the valid range [0, 1]
        transparency = max(0, min(1, transparency))

        return transparency

    def _transparency_to_magnitude_degradation(self, transparency):
        """
        The magnitude degradation based on the atmospheric transparency.

        Used by: magnitude degradation

        Args:
        - transparency: In the range of 0 to 1.

        Returns:
        - magniture_degradation: In magnitude.
        """

        if transparency == 0:
            return float(MAG_DEGRATION_MAX)
        magnitude_degradation = -MAG_DEGRATION_MAX * math.log(transparency)
        magnitude_degradation = max(0, min(MAG_DEGRATION_MAX, magnitude_degradation))

        return magnitude_degradation


class AstronomicalRoutines:
    """Calculate different astronomical objects."""

    def __init__(
        self,
        location_data,
        forecast_time=None,
    ):
        self._location_data = location_data
        self._test_mode = False

        _LOGGER.debug("Timezone: %s", self._location_data.timezone_info)
        if forecast_time is None:
            self._forecast_time = datetime.now(UTC).replace(tzinfo=UTC)
        else:
            self._forecast_time = forecast_time.replace(tzinfo=UTC)
            self._test_mode = True

        self._sun_observer = None
        self._sun_observer_nautical = None
        self._sun_observer_astro = None
        self._moon_observer = None
        self._sun = None
        self._moon = None
        self._sun_data = {}
        self._moon_data = {}
        self._darkness_data = {}

        # Internal only
        self._sun_previous_rising_astro = None
        self._sun_previous_setting_astro = None

    def _test_data(self, data, keys) -> bool:
        """Test that specific values in a dictionary are not None"""

        if not all(data[key] is not None for key in keys):
            return False
        return True

    def utc_to_local_diff(self):
        """Returns the UTC Offset."""

        # Get the current time in the specified timezone
        now = datetime.now(ZoneInfo(self._location_data.timezone_info))

        # Get the offset in seconds
        offset_seconds = now.utcoffset().total_seconds()

        # Convert the offset to hours
        return offset_seconds / 3600

    async def time_shift(self) -> int:
        """Returns the time_shift to UTC."""

        return int(self.utc_to_local_diff() * 3600)

    async def need_update(self, forecast_time=None):
        """Update Sun and Moon."""

        if forecast_time is not None:
            self._forecast_time = forecast_time.replace(tzinfo=UTC)

        _LOGGER.debug("Astronomical calculations updating")
        self._calculate_sun()
        self._calculate_moon()

    #
    # Observers
    #
    def _get_sun_observer(self, below_horizon=ASTRONOMICAL_DUSK_DAWN) -> ephem.Observer:
        """Retrieves the ephem sun observer for the current location."""

        observer = ephem.Observer()
        observer.lon = str(self._location_data.longitude)  # * degree
        observer.lat = str(self._location_data.latitude)  # * degree
        observer.elevation = self._location_data.elevation
        observer.horizon = below_horizon * degree
        observer.pressure = 0
        observer.epoch = datetime.now().strftime("%Y/%m/%d")
        return observer

    def _get_moon_observer(self) -> ephem.Observer:
        """Retrieves the ephem mon observer for the current location."""

        observer = ephem.Observer()
        observer.lon = str(self._location_data.longitude)  # * degree
        observer.lat = str(self._location_data.latitude)  # * degree
        observer.elevation = self._location_data.elevation
        # Naval Observatory Risings and Settings
        # Set horizon to minus 34 arcminutes
        # https://aa.usno.navy.mil/data/RS_OneDay
        observer.horizon = "-0:34"
        observer.pressure = 0
        observer.epoch = datetime.now().strftime("%Y/%m/%d")
        return observer

    # #########################################################################
    # Sun
    # #########################################################################
    async def sun_data(self) -> dict:
        """Returns sun data."""

        if self._sun_data is not None:
            try:
                return SunModel(**self._sun_data)
            except ValidationError as ve:
                _LOGGER.error(f"Failed to parse Sun data: {self._sun_data}")
                _LOGGER.error(ve)
                return None
        return None

    async def sun_next_rising(self) -> datetime:
        """Returns sun next rising."""

        if (
            not self._test_data(
                self._sun_data,
                ["next_rising_astro", "next_rising_nautical", "next_rising_civil"],
            )
            or self._forecast_time > self._sun_data["next_rising_astro"]
            or self._forecast_time > self._sun_data["next_rising_nautical"]
            or self._forecast_time > self._sun_data["next_rising_civil"]
        ):
            _LOGGER.debug("Astronomical calculations updating sun_next_rising")
            self._calculate_sun()

        if self._sun_data.get("next_rising_astro", None) is not None:
            return self._sun_data["next_rising_astro"]
        if self._sun_data.get("next_rising_nautical", None) is not None:
            return self._sun_data["next_rising_nautical"]
        if self._sun_data.get("next_rising_civil", None) is not None:
            return self._sun_data["next_rising_civil"]

    async def sun_next_setting(self) -> datetime:
        """Returns sun next setting."""

        if (
            not self._test_data(
                self._sun_data,
                ["next_rising_astro", "next_rising_nautical", "next_rising_civil"],
            )
            or self._forecast_time > self._sun_data["next_setting_astro"]
            or self._forecast_time > self._sun_data["next_setting_nautical"]
            or self._forecast_time > self._sun_data["next_setting_civil"]
        ):
            _LOGGER.debug("Astronomical calculations updating sun_next_setting")
            self._calculate_sun()

        if self._sun_data.get("next_setting_astro", None) is not None:
            return self._sun_data["next_setting_astro"]
        if self._sun_data.get("next_setting_nautical", None) is not None:
            return self._sun_data["next_setting_nautical"]
        if self._sun_data.get("next_setting_civil", None) is not None:
            return self._sun_data["next_setting_civil"]

    def _calculate_sun(self):
        """Calculates sun risings and settings."""

        if self._sun_observer is None:
            self._sun_observer = self._get_sun_observer(CIVIL_DUSK_DAWN)
        if self._sun_observer_nautical is None:
            self._sun_observer_nautical = self._get_sun_observer(NAUTICAL_DUSK_DAWN)
        if self._sun_observer_astro is None:
            self._sun_observer_astro = self._get_sun_observer(ASTRONOMICAL_DUSK_DAWN)
        if self._sun is None:
            self._sun = ephem.Sun()

        self._calculate_sun_civil()
        self._calculate_sun_nautical()
        self._calculate_sun_astro()
        self._calculate_sun_altaz()

    def _calculate_sun_civil(self):
        # Rise and Setting (Civil)
        try:
            self._sun_data["next_rising_civil"] = (
                self._sun_observer.next_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next rising
            start = self._sun_observer.date.datetime()
            end = self._sun_observer.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer.date = timestamp
                try:
                    self._sun_data["next_rising_civil"] = self._sun_observer.next_rising(
                        ephem.Sun(), use_center=True
                    ).datetime()
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_data["next_setting_civil"] = (
                self._sun_observer.next_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next setting
            start = self._sun_observer.date.datetime()
            end = self._sun_observer.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer.date = timestamp
                try:
                    self._sun_data["next_setting_civil"] = self._sun_observer.next_setting(
                        ephem.Sun(), use_center=True
                    ).datetime()
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

    def _calculate_sun_nautical(self):
        # Rise and Setting (Nautical)
        self._sun_observer_nautical.date = self._forecast_time
        self._sun.compute(self._sun_observer_nautical)

        try:
            self._sun_data["next_rising_nautical"] = (
                self._sun_observer_nautical.next_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical rising
            start = self._sun_observer_nautical.date.datetime()
            end = self._sun_observer_nautical.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_nautical.date = timestamp
                try:
                    self._sun_data["next_rising_nautical"] = (
                        self._sun_observer_nautical.next_rising(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_data["next_setting_nautical"] = (
                self._sun_observer_nautical.next_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical setting
            start = self._sun_observer_nautical.date.datetime()
            end = self._sun_observer_nautical.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_nautical.date = timestamp
                try:
                    self._sun_data["next_setting_nautical"] = (
                        self._sun_observer_nautical.next_setting(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

    def _calculate_sun_astro(self):
        # Rise and Setting (Astronomical)
        self._sun_observer_astro.date = self._forecast_time
        self._sun.compute(self._sun_observer_astro)

        try:
            self._sun_data["next_rising_astro"] = (
                self._sun_observer_astro.next_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical rising
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_data["next_rising_astro"] = (
                        self._sun_observer_astro.next_rising(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_data["previous_rising_astro"] = (
                self._sun_observer_astro.previous_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the previous astronomical rising
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() - timedelta(days=365)
            timestamp = start
            while timestamp > end:
                timestamp -= timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_data["previous_rising_astro"] = (
                        self._sun_observer_astro.previous_rising(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_data["next_setting_astro"] = (
                self._sun_observer_astro.next_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical setting
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_data["next_setting_astro"] = (
                        self._sun_observer_astro.next_setting(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_data["previous_setting_astro"] = (
                self._sun_observer_astro.previous_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the previous astronomical setting
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() - timedelta(days=365)
            timestamp = start
            while timestamp > end:
                timestamp -= timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_data["previous_setting_astro"] = (
                        self._sun_observer_astro.previous_setting(ephem.Sun(), use_center=True)
                        .datetime()
                        .replace(tzinfo=UTC)
                    )
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

    def _calculate_sun_altaz(self):
        """Calculates sun altitude and azimuth."""

        if self._sun_observer is None:
            self._sun_observer = self._get_sun_observer(CIVIL_DUSK_DAWN)
        if self._sun is None:
            self._sun = ephem.Sun()

        self._sun_observer.date = self._forecast_time
        self._sun.compute(self._sun_observer)

        # Sun Altitude
        self._sun_data["altitude"] = deg(float(self._sun.alt))

        # Sun Azimuth
        self._sun_data["azimuth"] = deg(float(self._sun.az))

    # #########################################################################
    # Moon
    # #########################################################################
    async def moon_data(self) -> dict:
        """Returns moon data."""

        if self._moon_data is not None:
            try:
                return MoonModel(**self._moon_data)
            except ValidationError as ve:
                _LOGGER.error(f"Failed to parse Moon data: {self._moon_data}")
                _LOGGER.error(ve)
                return None
        return None

    def _calculate_moon(self):
        """Calculates moon rising and setting."""

        if self._moon_observer is None:
            self._moon_observer = self._get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        # Rise and Setting
        self._moon_observer.date = self._forecast_time
        self._moon.compute(self._moon_observer)

        try:
            self._moon_data["next_rising"] = (
                self._moon_observer.next_rising(ephem.Moon()).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            pass

        try:
            self._moon_data["next_setting"] = (
                self._moon_observer.next_setting(ephem.Moon()).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            pass

        try:
            self._moon_data["previous_rising"] = (
                self._moon_observer.previous_rising(ephem.Moon()).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            pass

        try:
            self._moon_data["previous_setting"] = (
                self._moon_observer.previous_setting(ephem.Moon()).datetime().replace(tzinfo=UTC)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            pass

        # self._moon_observer.date = self._forecast_time + timedelta(days=1)
        # self._moon.compute(self._moon_observer)

        # try:
        #     self._moon_day_after_next_rising = self._moon_observer.next_rising(ephem.Moon()).datetime().replace(tzinfo=UTC)
        # except (ephem.AlwaysUpError, ephem.NeverUpError):
        #     pass

        # try:
        #     self._moon_day_after_next_setting = self._moon_observer.next_setting(ephem.Moon()).datetime().replace(tzinfo=UTC)
        # except (ephem.AlwaysUpError, ephem.NeverUpError):
        #     pass

        # Next new Moon
        self._moon_data["next_new_moon"] = ephem.next_new_moon(self._forecast_time).datetime().replace(tzinfo=UTC)

        # Next full Moon
        self._moon_data["next_full_moon"] = ephem.next_full_moon(self._forecast_time).datetime().replace(tzinfo=UTC)

        # Moon phase
        self._moon_data["phase"] = self._moon.phase

        self._calculate_moon_altaz()
        self._calculate_moon_distance_size()

    def _calculate_moon_altaz(self):
        """Calculates moon altitude and azimuth."""

        if self._moon_observer is None:
            self._moon_observer = self._get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        self._moon_observer.date = self._forecast_time
        self._moon.compute(self._moon_observer)

        # Moon Altitude
        self._moon_data["altitude"] = deg(float(self._moon.alt))

        # Moon Azimuth
        self._moon_data["azimuth"] = deg(float(self._moon.az))

    def _calculate_moon_distance_size(self):
        """Calculate moon distance and relative size"""
        if self._moon_observer is None:
            self._moon_observer = self._get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        # Get the distance in Earth radii
        self._moon_data["distance"] = self._moon.earth_distance  # in AU (Astronomical Units)

        # Convert to kilometers
        self._moon_data["distance_km"] = self._moon_data["distance"] * 149597870.7  # 1 AU = 149597870.7 km

        # Get the Moon's angular size (in degrees)
        self._moon_data["angular_size"] = self._moon.radius * 2 * 57.29578  # 180 / pi

        # Average distance and angular size for comparison
        self._moon_data["avg_distance_km"] = 384400  # Average distance of the Moon from Earth in km
        self._moon_data["avg_angular_size"] = 0.5181  # Average angular size in degrees

        # Relative distance and size compared to average
        self._moon_data["relative_distance"] = self._moon_data["distance_km"] / self._moon_data["avg_distance_km"]
        self._moon_data["relative_size"] = self._moon_data["angular_size"] / self._moon_data["avg_angular_size"]

    # #########################################################################
    # Darkness
    # #########################################################################
    async def darkness_data(self) -> dict:
        """Returns darkness data."""

        if not self._test_data(self._moon_data, ["next_setting", "next_rising", "previous_setting"]):
            _LOGGER.debug("DSD: moon data missing")
            return False

        if not self._test_data(
            self._sun_data,
            ["previous_setting_astro", "next_setting_astro", "next_rising_astro"],
        ):
            _LOGGER.debug("DSD: sun data missing")
            return False

        self._darkness_data["deep_sky_darkness_moon_rises"] = self._deep_sky_darkness_moon_rises()
        self._darkness_data["deep_sky_darkness_moon_sets"] = self._deep_sky_darkness_moon_sets()
        self._darkness_data["deep_sky_darkness_moon_always_up"] = self._deep_sky_darkness_moon_always_up()
        self._darkness_data["deep_sky_darkness_moon_always_down"] = self._deep_sky_darkness_moon_always_down()
        self._darkness_data["deep_sky_darkness"] = self._deep_sky_darkness()

        try:
            return DarknessModel(**self._darkness_data)
        except ValidationError as ve:
            _LOGGER.error(f"Failed to parse darkness data: {self._darkness_data}")
            _LOGGER.error(ve)
            return None

    async def night_duration_astronomical(self) -> float:
        """Returns the remaining timespan of astronomical darkness."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        return (self._sun_data["next_rising_astro"] - start_timestamp).total_seconds()

    def _deep_sky_darkness(self) -> float:
        """Returns the remaining timespan of deep sky darkness."""

        dsd = timedelta(0)

        if self._astronomical_darkness():
            _LOGGER.debug("DSD: In astronomical darkness")
            if self._deep_sky_darkness_moon_rises():
                dsd = self._moon_data("next_rising") - self._forecast_time
                _LOGGER.debug(f"DSD: Sun down, Moon rises {dsd}")

            if self._deep_sky_darkness_moon_sets():
                if self._moon_down():
                    dsd = self._sun_data["next_rising_astro"] - self._forecast_time
                    _LOGGER.debug(f"DSD: Sun down, Moon is down {dsd}")
                else:
                    dsd = self._sun_data["next_rising_astro"] - self._moon_data["next_setting"]
                    _LOGGER.debug(f"DSD: Sun down, Moon sets {dsd}")

            if self._deep_sky_darkness_moon_always_down():
                dsd = self._sun_data["next_rising_astro"] - self._forecast_time
                _LOGGER.debug(f"DSD: Moon always down {dsd}")
            else:
                _LOGGER.debug(f"DSD: Moon NOT always down {dsd}")

        if not self._astronomical_darkness():
            _LOGGER.debug("DSD: At sunlight")
            if self._deep_sky_darkness_moon_rises():
                dsd = self._moon_data["next_rising"] - self._sun_data["next_setting_astro"]
                _LOGGER.debug(f"DSD: Sun up, Moon rises {dsd}")

            if self._deep_sky_darkness_moon_sets():
                dsd = self._sun_data["next_rising_astro"] - self._moon_data["next_setting"]
                _LOGGER.debug(f"DSD: Sun up, Moon sets {dsd}")

            if self._deep_sky_darkness_moon_always_down():
                dsd = self._sun_data["next_rising_astro"] - self._sun_data["next_setting_astro"]
                _LOGGER.debug(f"DSD: Sun up, Moon down {dsd}")

        if self._deep_sky_darkness_moon_always_up():
            dsd = timedelta(0)
            _LOGGER.debug(f"DSD: Moon always up {dsd}")
        else:
            _LOGGER.debug(f"DSD: Moon NOT always up {dsd}")

        return dsd.total_seconds()

    def _astronomical_darkness(self) -> bool:
        """Returns true during astronomical night."""

        if self._sun_data["next_setting_astro"] > self._sun_data["next_rising_astro"]:
            return True
        return False

    def _moon_down(self) -> bool:
        """Returns true while moon is set-"""

        if self._moon_data["next_setting"] > self._moon_data["next_rising"]:
            return True
        return False

    def _deep_sky_darkness_moon_rises(self) -> bool:
        """Returns true if moon rises during astronomical night."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        if (
            self._moon_data["next_rising"] > start_timestamp
            and self._moon_data["next_rising"] < self._sun_data["next_rising_astro"]
        ):
            _LOGGER.debug("DSD: Moon rises during astronomical night")
            return True
        return False

    def _deep_sky_darkness_moon_sets(self) -> bool:
        """Returns true if moon sets during astronomical night."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        # Did Moon already set in darkness?
        if self._moon_down() and self._astronomical_darkness():
            start_timestamp_moon = self._moon_data["previous_setting"]
        else:
            start_timestamp_moon = self._moon_data["next_setting"]

        if start_timestamp_moon > start_timestamp and start_timestamp_moon < self._sun_data["next_rising_astro"]:
            _LOGGER.debug("DSD: Moon sets during astronomical night")
            return True
        return False

    def _deep_sky_darkness_moon_always_up(self) -> bool:
        """Returns true if moon is up during astronomical night."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        if (
            self._moon_data["next_rising"] < start_timestamp
            and self._moon_data["next_setting"] > self._sun_data["next_rising_astro"]
        ):
            _LOGGER.debug("DSD: Moon is up during astronomical night")
            return True
        return False

    def _deep_sky_darkness_moon_always_down(self) -> bool:
        """Returns true if moon is down during astronomical night."""

        start_timestamp = None

        # Are we already in darkness?
        if self._astronomical_darkness():
            start_timestamp = self._sun_data["previous_setting_astro"]
        else:
            start_timestamp = self._sun_data["next_setting_astro"]

        if (
            self._moon_data["previous_setting"] < start_timestamp
            and self._moon_data["next_rising"] > self._sun_data["next_rising_astro"]
        ):
            _LOGGER.debug("DSD: Moon is down during astronomical night")
            return True
        return False



"""Define a client to interact with 7Timer."""

import asyncio
import json
import logging
import math
import os.path
from datetime import UTC, datetime, timedelta
from json.decoder import JSONDecodeError
from typing import Optional

import aiofiles
from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError
from pydantic import ValidationError

from pyastroweatherio.const import (
    BASE_URL_MET,
    BASE_URL_SEVENTIMER,
    DEFAULT_CACHE_TIMEOUT,
    DEFAULT_CONDITION_CALM_WEIGHT,
    DEFAULT_CONDITION_CLOUDCOVER_HIGH_WEAKENING,
    DEFAULT_CONDITION_CLOUDCOVER_LOW_WEAKENING,
    DEFAULT_CONDITION_CLOUDCOVER_MEDIUM_WEAKENING,
    DEFAULT_CONDITION_CLOUDCOVER_WEIGHT,
    DEFAULT_CONDITION_SEEING_WEIGHT,
    DEFAULT_CONDITION_TRANSPARENCY_WEIGHT,
    DEFAULT_ELEVATION,
    DEFAULT_LATITUDE,
    DEFAULT_LONGITUDE,
    DEFAULT_TIMEOUT,
    DEFAULT_TIMEZONE,
    FORECAST_TYPE_HOURLY,
    HEADERS,
    LIFTED_INDEX_7TIMER_MAPPING,
    MAG_DEGRATION_MAX,
    SEEING,
    SEEING_MAX,
    TRANSPARENCY,
    WIND10M_MAX,
)
from pyastroweatherio.dataclasses import (
    BODIESUpTonight,
    COMETSUpTonight,
    DSOUpTonight,
    ForecastData,
    LocationData,
    NightlyConditionsData,
)
from pyastroweatherio.errors import RequestError
from pyastroweatherio.helper_functions import (
    AstronomicalRoutines,
    AtmosphericRoutines,
    ConversionFunctions,
)
from pyastroweatherio.models import (
    AtmosphereModel,
    ConditionModel,
    LocationModel,
    DSOUpTonightModel,
    BODIESUpTonightModel,
    COMETSUpTonightModel,
)

_LOGGER = logging.getLogger(__name__)
_NOT_AVAILABLE = -9999


class AstroWeather:
    """AstroWeather Communication Client."""

    def __init__(
        self,
        session: Optional[ClientSession] = None,
        latitude=DEFAULT_LATITUDE,
        longitude=DEFAULT_LONGITUDE,
        elevation=DEFAULT_ELEVATION,
        timezone_info=DEFAULT_TIMEZONE,
        cloudcover_weight=DEFAULT_CONDITION_CLOUDCOVER_WEIGHT,
        cloudcover_high_weakening=DEFAULT_CONDITION_CLOUDCOVER_HIGH_WEAKENING,
        cloudcover_medium_weakening=DEFAULT_CONDITION_CLOUDCOVER_MEDIUM_WEAKENING,
        cloudcover_low_weakening=DEFAULT_CONDITION_CLOUDCOVER_LOW_WEAKENING,
        seeing_weight=DEFAULT_CONDITION_SEEING_WEIGHT,
        transparency_weight=DEFAULT_CONDITION_TRANSPARENCY_WEIGHT,
        calm_weight=DEFAULT_CONDITION_CALM_WEIGHT,
        uptonight_path="/conf/www",
        test_datetime=None,
        experimental_features=False,
    ):
        self._session: ClientSession = session
        self._location_data = self._get_location(
            latitude,
            longitude,
            elevation,
            timezone_info,
        )
        self._timezone_info = timezone_info
        self._weather_data_seventimer = []
        self._weather_data_seventimer_init = ""
        self._weather_data_metno = []
        self._weather_data_metno_init = ""
        self._weather_data_uptonight = {}
        self._weather_data_uptonight_bodies = {}
        self._weather_data_uptonight_comets = {}
        self._weather_data_seventimer_timestamp = datetime.now() - timedelta(seconds=(DEFAULT_CACHE_TIMEOUT + 1))
        self._weather_data_metno_timestamp = datetime.now() - timedelta(seconds=(DEFAULT_CACHE_TIMEOUT + 1))
        self._data_uptonight_timestamp = datetime.now() - timedelta(seconds=(DEFAULT_CACHE_TIMEOUT + 1))
        self._cloudcover_weight = cloudcover_weight
        self._cloudcover_high_weakening = cloudcover_high_weakening
        self._cloudcover_medium_weakening = cloudcover_medium_weakening
        self._cloudcover_low_weakening = cloudcover_low_weakening
        self._seeing_weight = seeing_weight
        self._transparency_weight = transparency_weight
        self._calm_weight = calm_weight
        self._uptonight_path = uptonight_path
        self._test_datetime = test_datetime
        self._experimental_features = experimental_features

        self._forecast_data = None

        self.req = session

        # Astro Routines
        self._astro_routines = AstronomicalRoutines(
            self._location_data,
            self._test_datetime,
        )

        self._atmosphere = AtmosphericRoutines()

        # Testing
        self._test_mode = False
        if self._test_datetime is not None:
            self._test_mode = True

    # #########################################################################
    # Public functions
    # #########################################################################
    async def get_location_data(
        self,
    ) -> None:
        """Returns station Weather Forecast."""

        return await self._get_location_data()

    async def get_hourly_forecast(self) -> None:
        """Returns hourly Weather Forecast."""

        return await self._get_forecast_data(FORECAST_TYPE_HOURLY, 72)

    async def get_deepsky_forecast(self) -> None:
        """Returns Deep Sky Forecast."""

        return await self._get_deepsky_forecast()

    # #########################################################################
    # Private functions
    # #########################################################################
    def _get_location(
        self,
        latitude,
        longitude,
        elevation,
        timezone_info,
    ) -> LocationModel | None:
        location = {
            "latitude": latitude,
            "longitude": longitude,
            "elevation": elevation,
            "timezone_info": timezone_info,
        }

        try:
            return LocationModel(**location)
        except ValidationError as ve:
            _LOGGER.error(f"Failed to parse location data: {location}")
            _LOGGER.error(ve)
            return None

    async def _get_atmosphere(self, details_seventimer, details_metno) -> AtmosphereModel | None:
        atmosphere = {}

        seeing = 0
        if details_seventimer["seeing"] == _NOT_AVAILABLE:
            seeing = await self._atmosphere.calculate_seeing(
                temperature=details_metno.get("air_temperature"),
                humidity=details_metno.get("relative_humidity"),
                dew_point_temperature=details_metno.get("dew_point_temperature"),
                wind_speed=details_metno.get("wind_speed"),
                cloud_cover=details_metno.get("cloud_area_fraction"),
                altitude=self._location_data.elevation,
                air_pressure_at_sea_level=details_metno.get("air_pressure_at_sea_level"),
            )
        else:
            seeing = SEEING[max(0, min(7, int(details_seventimer["seeing"] - 1)))]

        transparency = 0
        if details_seventimer["transparency"] == _NOT_AVAILABLE:
            transparency = await self._atmosphere.magnitude_degradation(
                temperature=details_metno.get("air_temperature"),
                humidity=details_metno.get("relative_humidity"),
                dew_point_temperature=details_metno.get("dew_point_temperature"),
                wind_speed=details_metno.get("wind_speed"),
                cloud_cover=details_metno.get("cloud_area_fraction"),
                altitude=self._location_data.elevation,
                air_pressure_at_sea_level=details_metno.get("air_pressure_at_sea_level"),
            )
        else:
            transparency = TRANSPARENCY[max(0, min(7, int(details_seventimer["transparency"] - 1)))]

        lifted_index = 0
        if details_seventimer["lifted_index"] == _NOT_AVAILABLE:
            lifted_index = await self._atmosphere.calculate_lifted_index(
                temperature=details_metno.get("air_temperature"),
                dew_point_temperature=details_metno.get("dew_point_temperature"),
                altitude=self._location_data.elevation,
                air_pressure_at_sea_level=details_metno.get("air_pressure_at_sea_level"),
            )
        else:
            lifted_index = LIFTED_INDEX_7TIMER_MAPPING[details_seventimer["lifted_index"]]

        atmosphere = {
            "seeing": seeing,
            "transparency": transparency,
            "lifted_index": lifted_index,
        }

        try:
            return AtmosphereModel(**atmosphere)
        except ValidationError as ve:
            _LOGGER.error(f"Failed to parse atmosphere data: {atmosphere}")
            _LOGGER.error(ve)
            return None

    async def _get_condition(
        self,
        details_metno,
        details_metno_next_1_hours,
        details_metno_next_6_hours,
        seeing,
        transparency,
        lifted_index,
    ) -> ConditionModel | None:
        condition = {
            "cloudcover": details_metno.get("cloud_area_fraction"),
            "cloud_area_fraction": details_metno.get("cloud_area_fraction"),
            "cloud_area_fraction_high": details_metno.get("cloud_area_fraction_high"),
            "cloud_area_fraction_low": details_metno.get("cloud_area_fraction_low"),
            "cloud_area_fraction_medium": details_metno.get("cloud_area_fraction_medium"),
            "fog_area_fraction": details_metno.get("fog_area_fraction"),
            "seeing": seeing,
            "transparency": transparency,
            "lifted_index": lifted_index,
            "condition_percentage": await self._calc_condition_percentage(
                details_metno.get("cloud_area_fraction_high"),
                details_metno.get("cloud_area_fraction_medium"),
                details_metno.get("cloud_area_fraction_low"),
                seeing,
                transparency,
                details_metno.get("wind_speed"),
                details_metno_next_1_hours.get("details", {}).get("precipitation_amount"),
            ),
            "rh2m": details_metno.get("relative_humidity"),
            "wind_speed": details_metno.get("wind_speed"),
            "wind_from_direction": details_metno.get("wind_from_direction"),
            "temp2m": details_metno.get("air_temperature"),
            "dewpoint2m": details_metno.get("dew_point_temperature"),
            "weather": details_metno_next_1_hours.get("summary", {}).get("symbol_code"),
            "weather6": details_metno_next_6_hours.get("summary", {}).get("symbol_code"),
            "precipitation_amount": details_metno_next_1_hours.get("details", {}).get("precipitation_amount"),
            "precipitation_amount6": details_metno_next_6_hours.get("details", {}).get("precipitation_amount"),
        }

        try:
            return ConditionModel(**condition)
        except ValidationError as ve:
            _LOGGER.error(f"Failed to parse condition data: {condition}")
            _LOGGER.error(ve)
            return None

    async def _calc_condition_percentage(
        self,
        cloudcover_high,
        cloudcover_medium,
        cloudcover_low,
        seeing,
        transparency,
        wind_speed,
        precipitation_amount,
    ):
        """Return condition based on cloud cover, seeing, transparency, and wind speed."""

        if not all(
            v is not None
            for v in [
                cloudcover_high,
                cloudcover_medium,
                cloudcover_low,
                seeing,
                transparency,
                wind_speed,
                precipitation_amount,
            ]
        ):
            return None

        # Seeing is something in between 0 and 2.5 arcsecs
        seeing = seeing * 100 / SEEING_MAX  # arcsecs up to 2.5
        # transparency = int(transparency * 40)  # mag degration up to 2.5
        transparency = transparency * 100 / MAG_DEGRATION_MAX  # mag degration up to MAG_DEGRATION_MAX
        # Wind speed is something in between 0 and 16.5 m/s
        if wind_speed > WIND10M_MAX:
            wind_speed = WIND10M_MAX
        wind_speed_value = int(wind_speed * (100 / WIND10M_MAX))  # m/s up to 16.5

        cloudcover = []
        cloudcover.append(cloudcover_high * self._cloudcover_high_weakening)
        cloudcover.append(cloudcover_medium * self._cloudcover_medium_weakening)
        cloudcover.append(cloudcover_low * self._cloudcover_low_weakening)

        condition = int(
            100
            - (
                self._cloudcover_weight * max(cloudcover)
                + self._seeing_weight * seeing
                + self._transparency_weight * transparency
                + self._calm_weight * wind_speed_value
            )
            / (self._cloudcover_weight + self._seeing_weight + self._transparency_weight + self._calm_weight)
            - precipitation_amount * 100
        )

        # Ensure condition is within the valid range [0, 1]
        condition = max(0, min(100, condition))

        return condition

    # #########################################################################
    # Data for AstroWeather
    # #########################################################################
    async def _get_location_data(self) -> None:
        """Return Forecast data"""

        cnv = ConversionFunctions()
        items = []

        await self._retrieve_data_metno()
        await self._retrieve_data_seventimer()
        now = datetime.now(UTC).replace(tzinfo=None)

        if self._test_datetime is not None:
            await self._astro_routines.need_update()
        else:
            await self._astro_routines.need_update(forecast_time=now)

        forecast_time = now.replace(minute=0, second=0, microsecond=0)
        if self._test_datetime is not None:
            forecast_time = self._test_datetime.replace(minute=0, second=0, microsecond=0)
        _LOGGER.debug("Forecast time: %s", str(forecast_time))

        if len(self._weather_data_metno) == 0:
            _LOGGER.error("Met.no data not available")
            return []

        # Met.no
        metno_index = 0
        seventimer_index = 0

        # Met.no: Search for start index
        for metno_index, datapoint in enumerate(self._weather_data_metno):
            if forecast_time == datetime.strptime(datapoint.get("time"), "%Y-%m-%dT%H:%M:%SZ"):
                break
        if metno_index > len(self._weather_data_metno):
            _LOGGER.error("Met.no start index not found")
            return []
        _LOGGER.debug("Met.no start index: %s", str(metno_index))

        details_metno = self._weather_data_metno[metno_index].get("data", {}).get("instant", {}).get("details", {})
        details_metno_next_1_hours = self._weather_data_metno[metno_index].get("data", {}).get("next_1_hours", {})
        details_metno_next_6_hours = self._weather_data_metno[metno_index].get("data", {}).get("next_6_hours", {})

        # Break condition
        if details_metno_next_1_hours is None:
            # No more hourly data
            _LOGGER.debug(
                "No more hourly data at %s",
                self._weather_data_metno[metno_index].get("time", {}),
            )
            return None
        if details_metno_next_6_hours is None:
            # No more 6-hourly data
            _LOGGER.debug(
                "No more 6-hourly data at %s",
                self._weather_data_metno[metno_index].get("time", {}),
            )
            return None

        # 7Timer: Search for start index
        seventimer_init = await cnv.anchor_timestamp(self._weather_data_seventimer_init)
        for row7 in self._weather_data_seventimer:
            if seventimer_init + timedelta(hours=row7["timepoint"]) > now:
                break
            seventimer_index += 1
        if seventimer_index < len(self._weather_data_seventimer):
            _LOGGER.debug("7Timer start index: %s", str(seventimer_index))
            details_seventimer = self._weather_data_seventimer[seventimer_index]
        else:
            details_seventimer = {
                "timepoint": 0,
                "seeing": _NOT_AVAILABLE,
                "transparency": _NOT_AVAILABLE,
                "lifted_index": _NOT_AVAILABLE,
            }

        atmosphere_data = await self._get_atmosphere(details_seventimer, details_metno)

        item = {
            # seventimer_init is "init" of 7timer astro data
            "seventimer_init": seventimer_init,  # init
            # seventimer_timepoint is "timepoint" of 7timer astro data and defines the data for init + timepoint
            "seventimer_timepoint": details_seventimer["timepoint"],  # timepoint
            # Forecast_time is the actual datetime for the forecast data onwards in UTC
            # Corresponds to "time" in met data
            "forecast_time": forecast_time,  # timestamp
            # Time shift to UTC
            "time_shift": await self._astro_routines.time_shift(),
            # Remaining forecast data point in met.no data
            "forecast_length": (len(self._weather_data_metno) - metno_index),
            # Location
            "location_data": self._location_data,
            # Astronomical routines
            "sun_data": await self._astro_routines.sun_data(),
            "moon_data": await self._astro_routines.moon_data(),
            "darkness_data": await self._astro_routines.darkness_data(),
            "night_duration_astronomical": await self._astro_routines.night_duration_astronomical(),
            "deepsky_forecast": await self._get_deepsky_forecast(),
            # Met.no
            "condition_data": await self._get_condition(
                details_metno,
                details_metno_next_1_hours,
                details_metno_next_6_hours,
                atmosphere_data.seeing,
                atmosphere_data.transparency,
                atmosphere_data.lifted_index,
            ),
            # Uptonight objects
            "uptonight": await self._get_deepsky_objects(),
            "uptonight_bodies": await self._get_bodies(),
            "uptonight_comets": await self._get_comets(),
        }

        items.append(LocationData(item))

        return items

    async def _get_forecast_data(self, forecast_type, hours_to_show) -> None:
        """Return Forecast data for the Station."""

        cnv = ConversionFunctions()
        items = []

        await self._retrieve_data_metno()
        await self._retrieve_data_seventimer()
        now = datetime.now(UTC).replace(tzinfo=None)

        # Create items
        cnt = 0

        forecast_time = now.replace(minute=0, second=0, microsecond=0)
        if self._test_datetime is not None:
            forecast_time = self._test_datetime.replace(minute=0, second=0, microsecond=0)
        _LOGGER.debug("Forecast time: %s", str(forecast_time))

        # 7Timer: Search for start index
        seventimer_init = await cnv.anchor_timestamp(self._weather_data_seventimer_init)

        # Anchor timestamp
        init_ts = await cnv.anchor_timestamp(self._weather_data_seventimer_init)

        utc_to_local_diff = self._astro_routines.utc_to_local_diff()
        _LOGGER.debug("UTC to local diff: %s", str(utc_to_local_diff))
        # _LOGGER.debug("Forecast length 7timer: %s", str(len(self._weather_data_seventimer)))

        if len(self._weather_data_metno) == 0:
            _LOGGER.error("Met.no data not available")
            return []

        last_forecast_time = forecast_time
        for metno_index, datapoint in enumerate(self._weather_data_metno):
            if forecast_time > datetime.strptime(datapoint.get("time"), "%Y-%m-%dT%H:%M:%SZ"):
                continue

            if datetime.strptime(datapoint.get("time"), "%Y-%m-%dT%H:%M:%SZ") - last_forecast_time > timedelta(hours=1):
                break

            last_forecast_time = datetime.strptime(datapoint.get("time"), "%Y-%m-%dT%H:%M:%SZ")
            details_metno = datapoint.get("data", {}).get("instant", {}).get("details")

            if details_metno.get("cloud_area_fraction") is None:
                _LOGGER.error("Missing Met.no data")
                break

            details_seventimer = self._get_data_seventimer_timer(
                seventimer_init,
                datetime.strptime(datapoint.get("time"), "%Y-%m-%dT%H:%M:%SZ"),
            )
            details_metno_next_1_hours = self._weather_data_metno[metno_index].get("data", {}).get("next_1_hours")
            details_metno_next_6_hours = self._weather_data_metno[metno_index].get("data", {}).get("next_6_hours")

            # Break condition
            if details_metno_next_1_hours is None:
                # No more hourly data
                _LOGGER.debug(
                    "No more hourly data at %s",
                    self._weather_data_metno[metno_index].get("time", {}),
                )
                break
            if details_metno_next_6_hours is None:
                # No more 6-hourly data
                _LOGGER.debug(
                    "No more 6-hourly data at %s",
                    self._weather_data_metno[metno_index].get("time", {}),
                )
                break

            atmosphere_data = await self._get_atmosphere(details_seventimer, details_metno)

            item = {
                "seventimer_init": init_ts,
                "seventimer_timepoint": details_seventimer["timepoint"],
                "forecast_time": datetime.strptime(datapoint.get("time"), "%Y-%m-%dT%H:%M:%SZ"),
                "hour": datetime.strptime(datapoint.get("time"), "%Y-%m-%dT%H:%M:%SZ").hour,  # forecast_time.hour % 24,
                "condition_data": await self._get_condition(
                    details_metno,
                    details_metno_next_1_hours,
                    details_metno_next_6_hours,
                    atmosphere_data.seeing,
                    atmosphere_data.transparency,
                    atmosphere_data.lifted_index,
                ),
            }

            items.append(ForecastData(item))

            item["seventimer_timepoint"] = item["seventimer_timepoint"] + 1
            item["forecast_time"] = item["forecast_time"] + timedelta(hours=1)
            item["hour"] = item["hour"] + 1

            cnt += 1
            if cnt >= hours_to_show:
                break

        self._forecast_data = items

        _LOGGER.debug("Forceast Length: %s", str(len(items)))
        return items

    async def _get_deepsky_forecast(self):
        """Return Deepsky Forecast data."""

        cnv = ConversionFunctions()
        items = []

        if self._forecast_data is None:
            await self._get_forecast_data(FORECAST_TYPE_HOURLY, 72)

        # Anchor timestamp
        init_ts = await cnv.anchor_timestamp(self._weather_data_seventimer_init)

        utc_to_local_diff = self._astro_routines.utc_to_local_diff()
        _LOGGER.debug("UTC to local diff: %s", str(utc_to_local_diff))

        # if self._forecast_data is None:
        #     _LOGGER.error("Met.no forecast data not available")
        #     return []

        # Create forecast
        forecast_dayname = ""
        start_forecast_hour = 0
        start_weather = ""
        interval_points = []
        now = datetime.now(UTC).replace(tzinfo=None)

        if self._test_datetime is not None:
            await self._astro_routines.need_update()
        else:
            await self._astro_routines.need_update(forecast_time=now)

        sun_next_setting = await self._astro_routines.sun_next_setting()
        sun_next_rising = await self._astro_routines.sun_next_rising()
        night_duration_astronomical = await self._astro_routines.night_duration_astronomical()

        start_indexes = []
        # Find start index for two nights and store the indexes
        for index, details_forecast in enumerate(self._forecast_data):
            if details_forecast.forecast_time.hour % 24 == sun_next_rising.hour and len(start_indexes) == 0:
                start_indexes.append(0)
            if details_forecast.forecast_time.hour % 24 == sun_next_setting.hour:
                start_indexes.append(index)

        forecast_data_len = len(self._forecast_data)
        for day in range(0, len(start_indexes)):
            start_forecast_hour = 0
            start_weather = ""
            interval_points = []
            start_index = start_indexes[day]
            for index in range(
                start_index,
                start_index + int(math.floor(night_duration_astronomical / 3600) + 2),
            ):
                if index >= forecast_data_len:
                    _LOGGER.debug("No more forecast data")
                    break
                details_forecast = self._forecast_data[index]

                if len(interval_points) == 0:
                    forecast_dayname = details_forecast.forecast_time.strftime("%A")
                    start_forecast_hour = details_forecast.forecast_time.hour
                    start_weather = details_forecast.weather6
                    start_precipitation_amount6 = details_forecast.precipitation_amount6

                # Calculate Condition
                if len(interval_points) <= int(math.floor(night_duration_astronomical / 3600)):
                    interval_points.append(
                        await self._calc_condition_percentage(
                            details_forecast.cloud_area_fraction_high_percentage,
                            details_forecast.cloud_area_fraction_medium_percentage,
                            details_forecast.cloud_area_fraction_low_percentage,
                            details_forecast.seeing,
                            details_forecast.transparency,
                            details_forecast.wind10m_speed,
                            details_forecast.precipitation_amount,
                        )
                    )

                if details_forecast.forecast_time.hour == sun_next_rising.hour or index >= (forecast_data_len - 1):
                    item = {
                        "seventimer_init": init_ts,
                        "dayname": forecast_dayname,
                        "hour": start_forecast_hour,
                        "nightly_conditions": interval_points,
                        "weather": start_weather,
                        "precipitation_amount6": start_precipitation_amount6,
                    }
                    items.append(NightlyConditionsData(item))

                    conditions_numeric = ""
                    for condition in interval_points:
                        conditions_numeric += str(condition) + ", "
                    _LOGGER.debug(
                        "Nightly conditions day: %s, start hour: %s, nightly conditions: %s, weather: %s, conditions numeric: %s",
                        str(forecast_dayname),
                        str(start_forecast_hour),
                        str(len(interval_points)),
                        str(start_weather),
                        conditions_numeric,
                    )

                    # Test for end of astronomical night. Will get true if we're already at night.
                    if details_forecast.forecast_time.hour % 24 == sun_next_rising.hour:
                        break
            start_index += 24

        return items

    # #########################################################################
    # UpTonight
    # #########################################################################
    async def _get_deepsky_objects(self):
        """Return Deepsky Objects for today."""

        items = []

        await self._retrieve_data_uptonight()

        # Create list of deep sky objects
        if self._weather_data_uptonight is not None:
            dso_id = self._weather_data_uptonight.get("id", {})
            dso_target_name = self._weather_data_uptonight.get("target name", {})
            dso_type = self._weather_data_uptonight.get("type", {})
            dso_constellation = self._weather_data_uptonight.get("constellation", {})
            dso_size = self._weather_data_uptonight.get("size", {})
            dso_visual_magnitude = self._weather_data_uptonight.get("mag", {})
            dso_meridian_transit = self._weather_data_uptonight.get("meridian transit", {})
            dso_meridian_antitransit = self._weather_data_uptonight.get("meridian antitransit", {})
            dso_foto = self._weather_data_uptonight.get("foto", {})

            for row in range(len(dso_target_name)):
                dso_meridian_transit_local = dso_meridian_transit.get(str(row), "")
                if dso_meridian_transit_local != "":
                    dso_meridian_transit_utc = (
                        datetime.strptime(dso_meridian_transit_local, "%m/%d/%Y %H:%M:%S")
                        - timedelta(seconds=await self._astro_routines.time_shift())
                    ).replace(tzinfo=UTC)
                else:
                    dso_meridian_transit_utc = ""

                dso_meridian_antitransit_local = dso_meridian_antitransit.get(str(row), "")
                if dso_meridian_antitransit_local != "":
                    dso_meridian_antitransit_utc = (
                        datetime.strptime(dso_meridian_antitransit_local, "%m/%d/%Y %H:%M:%S")
                        - timedelta(seconds=await self._astro_routines.time_shift())
                    ).replace(tzinfo=UTC)
                else:
                    dso_meridian_antitransit_utc = ""

                item = {
                    "id": dso_id.get(str(row), ""),
                    "target_name": dso_target_name.get(str(row), ""),
                    "type": dso_type.get(str(row), ""),
                    "constellation": dso_constellation.get(str(row), ""),
                    "size": dso_size.get(str(row), ""),
                    "visual_magnitude": dso_visual_magnitude.get(str(row), ""),
                    "meridian_transit": dso_meridian_transit_utc,
                    "meridian_antitransit": dso_meridian_antitransit_utc,
                    "foto": dso_foto.get(str(row), 0),
                }
                try:
                    items.append(DSOUpTonightModel(**item))
                except ValidationError as ve:
                    _LOGGER.error(f"Failed to parse deep sky object data: {item}")
                    _LOGGER.error(ve)

            return items
        return None

    async def _get_bodies(self):
        """Return Bodies for today."""

        items = []

        await self._retrieve_data_uptonight()

        # Create list of bodies
        if self._weather_data_uptonight_bodies is not None:
            body_target_name = self._weather_data_uptonight_bodies.get("target name", {})
            body_max_altitude = self._weather_data_uptonight_bodies.get("max altitude", {})
            body_azimuth = self._weather_data_uptonight_bodies.get("azimuth", {})
            body_max_altitude_time = self._weather_data_uptonight_bodies.get("max altitude time", {})
            body_visual_magnitude = self._weather_data_uptonight_bodies.get("visual magnitude", {})
            body_meridian_transit = self._weather_data_uptonight_bodies.get("meridian transit", {})
            body_foto = self._weather_data_uptonight_bodies.get("foto", {})

            for row in range(len(body_target_name)):
                # UpTonight delivers the time in local time zone. here we need it in UTC
                body_max_altitude_time_local = body_max_altitude_time.get(str(row), "")
                body_max_altitude_time_utc = (
                    datetime.strptime(body_max_altitude_time_local, "%m/%d/%Y %H:%M:%S")
                    - timedelta(seconds=await self._astro_routines.time_shift())
                ).replace(tzinfo=UTC)

                body_meridian_transit_local = body_meridian_transit.get(str(row), "")
                if body_meridian_transit_local != "":
                    body_meridian_transit_utc = (
                        datetime.strptime(body_meridian_transit_local, "%m/%d/%Y %H:%M:%S")
                        - timedelta(seconds=await self._astro_routines.time_shift())
                    ).replace(tzinfo=UTC)
                else:
                    body_meridian_transit_utc = ""

                item = {
                    "target_name": body_target_name.get(str(row), ""),
                    "max_altitude": body_max_altitude.get(str(row), ""),
                    "azimuth": body_azimuth.get(str(row), ""),
                    "max_altitude_time": body_max_altitude_time_utc,
                    "visual_magnitude": body_visual_magnitude.get(str(row), ""),
                    "meridian_transit": body_meridian_transit_utc,
                    "foto": body_foto.get(str(row), 0),
                }
                try:
                    items.append(BODIESUpTonightModel(**item))
                except ValidationError as ve:
                    _LOGGER.error(f"Failed to parse bodies data: {item}")
                    _LOGGER.error(ve)

            return items
        return None

    async def _get_comets(self):
        """Return Comets for today."""

        items = []

        await self._retrieve_data_uptonight()

        # Create list of comets
        if self._weather_data_uptonight_comets is not None:
            comet_target_name = self._weather_data_uptonight_comets.get("target name", {})
            distance_au_earth = self._weather_data_uptonight_comets.get("distance earth au", {})
            distance_au_sun = self._weather_data_uptonight_comets.get("distance sun au", {})
            absolute_magnitude = self._weather_data_uptonight_comets.get("absolute magnitude", {})
            visual_magnitude = self._weather_data_uptonight_comets.get("visual magnitude", {})
            altitude = self._weather_data_uptonight_comets.get("altitude", {})
            azimuth = self._weather_data_uptonight_comets.get("azimuth", {})
            rise_time = self._weather_data_uptonight_comets.get("rise time", {})
            set_time = self._weather_data_uptonight_comets.get("set time", {})

            for row in range(len(comet_target_name)):
                # UpTonight delivers the time in local time zone. here we need it in UTC
                rise_time_local = rise_time.get(str(row), "")
                set_time_local = set_time.get(str(row), "")
                rise_time_local_utc = (
                    datetime.strptime(rise_time_local, "%m/%d/%Y %H:%M:%S")
                    - timedelta(seconds=await self._astro_routines.time_shift())
                ).replace(tzinfo=UTC)
                set_time_local_utc = (
                    datetime.strptime(set_time_local, "%m/%d/%Y %H:%M:%S")
                    - timedelta(seconds=await self._astro_routines.time_shift())
                ).replace(tzinfo=UTC)

                item = {
                    "designation": comet_target_name.get(str(row), ""),
                    "distance_au_earth": distance_au_earth.get(str(row), ""),
                    "distance_au_sun": distance_au_sun.get(str(row), ""),
                    "absolute_magnitude": absolute_magnitude.get(str(row), ""),
                    "visual_magnitude": visual_magnitude.get(str(row), ""),
                    "altitude": altitude.get(str(row), ""),
                    "azimuth": azimuth.get(str(row), ""),
                    "rise_time": rise_time_local_utc,
                    "set_time": set_time_local_utc,
                }
                try:
                    items.append(COMETSUpTonightModel(**item))
                except ValidationError as ve:
                    _LOGGER.error(f"Failed to parse comets data: {item}")
                    _LOGGER.error(ve)

            return items
        return None

    async def _retrieve_data_uptonight(self):
        """Retrieves current data from uptonight"""

        if ((datetime.now() - self._data_uptonight_timestamp).total_seconds()) > DEFAULT_CACHE_TIMEOUT:
            self._data_uptonight_timestamp = datetime.now()
            _LOGGER.debug("Updating data from uptonight")

            dataseries_dso = None
            dataseries_bodies = None
            dataseries_comets = None

            if os.path.exists(self._uptonight_path):
                if os.path.isfile(self._uptonight_path + "/uptonight-report.json"):
                    # _LOGGER.debug(f"Uptonight report found")
                    async with aiofiles.open(self._uptonight_path + "/uptonight-report.json", mode="r") as json_file:
                        contents = await json_file.read()
                    dataseries_dso = json.loads(contents)
                    _LOGGER.debug("Uptonight DSO imported")
                else:
                    _LOGGER.debug(f"File uptonight-report.json not found in {self._uptonight_path}")

                if os.path.isfile(self._uptonight_path + "/uptonight-bodies-report.json"):
                    # _LOGGER.debug(f"Uptonight report found")
                    async with aiofiles.open(
                        self._uptonight_path + "/uptonight-bodies-report.json", mode="r"
                    ) as json_file:
                        contents = await json_file.read()
                    dataseries_bodies = json.loads(contents)
                    _LOGGER.debug("Uptonight Bodies imported")
                else:
                    _LOGGER.debug(f"File uptonight-bodies-report.json not found in {self._uptonight_path}")

                if os.path.isfile(self._uptonight_path + "/uptonight-comets-report.json"):
                    # _LOGGER.debug(f"Uptonight report found")
                    async with aiofiles.open(
                        self._uptonight_path + "/uptonight-comets-report.json", mode="r"
                    ) as json_file:
                        contents = await json_file.read()
                    dataseries_comets = json.loads(contents)
                    _LOGGER.debug("Uptonight Comets imported")
                else:
                    _LOGGER.debug(f"File uptonight-comets-report.json not found in {self._uptonight_path}")
            else:
                _LOGGER.debug(
                    f"Path for UpTonight data not found. Current path: {self._uptonight_path}/uptonight-report.json"
                )

            self._weather_data_uptonight = dataseries_dso
            self._weather_data_uptonight_bodies = dataseries_bodies
            self._weather_data_uptonight_comets = dataseries_comets
        else:
            _LOGGER.debug("Using cached data for uptonight")

    # #########################################################################
    # 7Timer
    # #########################################################################
    async def _retrieve_data_seventimer(self):
        """Retrieves current data from 7timer."""

        if ((datetime.now() - self._weather_data_seventimer_timestamp).total_seconds()) > DEFAULT_CACHE_TIMEOUT:
            self._weather_data_seventimer_timestamp = datetime.now()
            _LOGGER.debug("Updating data from 7Timer")

            astro_dataseries = {}

            # Testing
            if not self._experimental_features:
                if self._test_mode:
                    if os.path.isfile("debug/astro.json"):
                        _LOGGER.debug("Reading 7Timer from file")
                        with open("debug/astro.json") as json_file:
                            astro_dataseries_json = json.load(json_file)
                            astro_dataseries = astro_dataseries_json.get("dataseries", {})
                            json_data_astro = {"init": astro_dataseries_json.get("init")}
                    else:
                        json_data_astro = await self._async_request_seventimer("astro", "get")
                        astro_dataseries = json_data_astro.get("dataseries", {})
                else:
                    json_data_astro = await self._async_request_seventimer("astro", "get")
                    astro_dataseries = json_data_astro.get("dataseries", {})

            if astro_dataseries != {} and not self._experimental_features:
                self._weather_data_seventimer = astro_dataseries
                self._weather_data_seventimer_init = json_data_astro.get("init")
            else:
                # Fake 7timer weather data if service is broken
                # This eliminates consideration of seeing, transparency, and lifted_index
                # and switches automatically to experimental functions.
                self._weather_data_seventimer = []
                for index in range(0, 20):
                    self._weather_data_seventimer.append(
                        {
                            "timepoint": index * 3,
                            "seeing": _NOT_AVAILABLE,
                            "transparency": _NOT_AVAILABLE,
                            "lifted_index": _NOT_AVAILABLE,
                        }
                    )
                self._weather_data_seventimer_init = datetime.now(UTC).replace(tzinfo=None).strftime("%Y%m%d%H")
        else:
            _LOGGER.debug("Using cached data for 7Timer")

    async def _async_request_seventimer(self, product="astro", method="get") -> dict:
        """Make a request against the 7timer API."""

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        # BASE_URL_SEVENTIMER = "https://www.7timer.info/bin/api.pl?lon=XX.XX&lat=YY.YY&product=astro&output=json"
        url = (
            str(f"{BASE_URL_SEVENTIMER}")
            + "?lon="
            + str("%.1f" % round(self._location_data.longitude, 2))
            + "&lat="
            + str("%.1f" % round(self._location_data.latitude, 2))
            + "&product="
            + str(product)
            + "&output=json"
        )
        try:
            _LOGGER.debug(f"Query url: {url}")
            async with session.request(method, url, headers=HEADERS, ssl=False) as resp:
                resp.raise_for_status()
                plain = str(await resp.text()).replace("\n", " ")
                data = json.loads(plain)

                if self._test_mode:
                    json_string = json.dumps(data)
                    with open("debug/" + product + ".json", "w") as outfile:
                        outfile.write(json_string)

                return data
        except JSONDecodeError as jsonerr:
            _LOGGER.error(f"JSON decode error, expecting value: {jsonerr}")
            return {}
        except asyncio.TimeoutError as tex:
            _LOGGER.error(f"Request to endpoint timed out: {tex}")
            return {}
        except ClientError as err:
            _LOGGER.error(f"Error requesting data: {err}")
            return {}

        finally:
            if not use_running_session:
                await session.close()

    def _get_data_seventimer_timer(self, anchor_timestamp, datetime):
        """Return 7Timer datapoint of interest."""

        seventimer_index = 0
        for index, row7 in enumerate(self._weather_data_seventimer):
            if anchor_timestamp + timedelta(hours=row7["timepoint"]) > datetime:
                seventimer_index = index - 1
                break
            # index += 1
        if seventimer_index >= len(self._weather_data_seventimer):
            return {
                "timepoint": row7["timepoint"],
                "seeing": _NOT_AVAILABLE,
                "transparency": _NOT_AVAILABLE,
                "lifted_index": _NOT_AVAILABLE,
            }
        else:
            return self._weather_data_seventimer[seventimer_index]

    # #########################################################################
    # Met.no
    # #########################################################################
    async def _retrieve_data_metno(self):
        """Retrieves current data from met."""

        if ((datetime.now() - self._weather_data_metno_timestamp).total_seconds()) > DEFAULT_CACHE_TIMEOUT:
            self._weather_data_metno_timestamp = datetime.now()
            _LOGGER.debug("Updating data from Met.no")

            dataseries = None

            # Testing
            if self._test_mode:
                if os.path.isfile("debug/met.json"):
                    _LOGGER.debug("Reading Met.no data from file")
                    with open("debug/met.json") as json_file:
                        dataseries = json.load(json_file).get("properties", {}).get("timeseries")
                else:
                    json_data_metno = await self._async_request_met("met", "get")
                    dataseries = json_data_metno.get("properties", {}).get("timeseries")
            else:
                json_data_metno = await self._async_request_met("met", "get")
                dataseries = json_data_metno.get("properties", {}).get("timeseries")

            if dataseries is not None:
                if len(dataseries) > 0:
                    self._weather_data_metno = dataseries
                    self._weather_data_metno_init = dataseries[0].get("time")
        else:
            _LOGGER.debug("Using cached data for Met.no")

    async def _async_request_met(self, product="met", method="get") -> dict:
        """Make a request against the 7timer API."""

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(
                timeout=ClientTimeout(total=DEFAULT_TIMEOUT),
            )

        # BASE_URL_MET = "https://api.met.no/weatherapi/locationforecast/2.0/complete?altitude=XX&lat=XX.XX&lon=XX.XX"
        url = (
            str(f"{BASE_URL_MET}")
            + "?lon="
            + str("%.1f" % round(self._location_data.longitude, 2))
            + "&lat="
            + str("%.1f" % round(self._location_data.latitude, 2))
            + "&altitude="
            + str(int(self._location_data.elevation))
        )

        try:
            _LOGGER.debug(f"Query url: {url}")
            async with session.request(method, url, headers=HEADERS) as resp:
                resp.raise_for_status()
                # plain = str(await resp.text()).replace("\n", " ")
                # data = json.loads(plain)
                data = await resp.json()

                if self._test_mode:
                    json_string = json.dumps(data)
                    with open("debug/" + product + ".json", "w") as outfile:
                        outfile.write(json_string)

                return data
        except JSONDecodeError as jsonerr:
            _LOGGER.error(f"JSON decode error, expecting value: {jsonerr}")
            return {}
        except asyncio.TimeoutError as tex:
            _LOGGER.error(f"Request to endpoint timed out: {tex}")
            return {}
        except ClientError as err:
            _LOGGER.error(f"Error requesting data: {err}")
            return {}

        finally:
            if not use_running_session:
                await session.close()



from datetime import datetime

from pydantic import BaseModel, Field

from pyastroweatherio.const import (
    DEFAULT_ELEVATION,
    DEFAULT_LATITUDE,
    DEFAULT_LONGITUDE,
    DEFAULT_TIMEZONE,
)


class LocationModel(BaseModel):
    """Model for the location"""

    latitude: float = Field(default=DEFAULT_LATITUDE)
    longitude: float = Field(default=DEFAULT_LONGITUDE)
    elevation: float = Field(default=DEFAULT_ELEVATION)
    timezone_info: str = Field(default=DEFAULT_TIMEZONE)


class ConditionModel(BaseModel):
    """Model for weather conditions"""

    cloudcover: float
    cloud_area_fraction: float
    cloud_area_fraction_high: float
    cloud_area_fraction_low: float
    cloud_area_fraction_medium: float
    fog_area_fraction: float
    seeing: float
    transparency: float
    lifted_index: float
    condition_percentage: int
    rh2m: float
    wind_speed: float
    wind_from_direction: float
    temp2m: float
    dewpoint2m: float
    weather: str
    weather6: str
    precipitation_amount: float
    precipitation_amount6: float


class AtmosphereModel(BaseModel):
    """Model for atmosperic conditions"""

    seeing: float
    transparency: float
    lifted_index: float


class SunModel(BaseModel):
    """Model for Sun data"""

    altitude: float
    azimuth: float
    next_rising_astro: datetime
    next_rising_civil: datetime
    next_rising_nautical: datetime
    next_setting_astro: datetime
    next_setting_civil: datetime
    next_setting_nautical: datetime
    previous_rising_astro: datetime
    previous_setting_astro: datetime


class MoonModel(BaseModel):
    """Model for Moon data"""

    altitude: float
    angular_size: float
    avg_angular_size: float
    avg_distance_km: float
    azimuth: float
    distance: float
    distance_km: float
    next_full_moon: datetime
    next_new_moon: datetime
    next_rising: datetime
    next_setting: datetime
    phase: float
    previous_rising: datetime
    previous_setting: datetime
    relative_distance: float
    relative_size: float


class DarknessModel(BaseModel):
    """Model for deep sky darkness"""

    deep_sky_darkness_moon_rises: bool
    deep_sky_darkness_moon_sets: bool
    deep_sky_darkness_moon_always_up: bool
    deep_sky_darkness_moon_always_down: bool
    deep_sky_darkness: float


class DSOUpTonightModel(BaseModel):
    """Model for DSO objects"""

    id: str
    target_name: str
    type: str
    constellation: str
    size: float
    visual_magnitude: float
    meridian_transit: datetime | str
    meridian_antitransit: datetime | str
    foto: float


class BODIESUpTonightModel(BaseModel):
    """Model for bodies"""

    target_name: str
    max_altitude: float
    azimuth: float
    max_altitude_time: datetime
    visual_magnitude: float
    meridian_transit: datetime | str
    foto: float


class COMETSUpTonightModel(BaseModel):
    """Model for comets"""

    designation: str
    distance_au_earth: float
    distance_au_sun: float
    absolute_magnitude: float
    visual_magnitude: float
    altitude: float
    azimuth: float
    rise_time: datetime
    set_time: datetime
